"""Preflight launcher that locates a healthy Python runtime before executing gloggur.

When ``scripts/gloggur`` is invoked in a fresh environment, it runs this module
to probe candidate interpreters (venv first, then system Python fallbacks) and
exec into the first one that can successfully import the gloggur package.  If
all probes fail, it emits a structured JSON or human-readable failure payload
with an actionable error code and remediation steps, then exits non-zero.

Exit codes (see ``EXIT_CODES``):
  2  missing_venv       - repository virtualenv missing and no healthy fallback
  3  missing_python     - no usable Python interpreter found on PATH
  4  missing_package    - interpreter found but gloggur dependencies absent
  5  broken_environment - other runtime/import failure
"""

from __future__ import annotations

import json
import os
import shutil
import subprocess
import sys
import time
from collections.abc import Sequence
from dataclasses import asdict, dataclass
from pathlib import Path

PROBE_MODULE_DEFAULT = "gloggur.cli.main"
VENV_EXEC_MODULE = "gloggur.cli.main"
SYSTEM_EXEC_MODULE = "gloggur"
CHECK_OPERATION = "preflight"
BOOTSTRAP_STAGE = "bootstrap"
BOOTSTRAP_STRICT_ENV = "GLOGGUR_BOOTSTRAP_STRICT"

EXIT_CODES = {
    "missing_venv": 2,
    "missing_python": 3,
    "missing_package": 4,
    "broken_environment": 5,
}


@dataclass
class CandidateProbe:
    """Result of probing a single Python interpreter candidate.

    ``exists`` indicates whether the interpreter binary was found on disk or
    PATH.  ``healthy`` indicates the interpreter could successfully import the
    required gloggur module.  ``reason`` and ``detail`` are populated on failure.
    """

    candidate_type: str
    interpreter: str
    module: str
    exists: bool
    healthy: bool
    reason: str | None = None
    detail: str | None = None
    returncode: int | None = None


@dataclass
class LaunchPlan:
    """Resolved execution plan produced by ``build_launch_plan``.

    When ``ready`` is ``True``, ``args``, ``interpreter``, and ``module`` are
    populated and the caller may safely exec into the interpreter.  When
    ``ready`` is ``False``, ``error_code``, ``message``, and ``remediation``
    describe the failure and no exec should be attempted.
    """

    ready: bool
    args: list[str]
    interpreter: str | None
    module: str | None
    candidate_type: str | None
    env: dict[str, str]
    repo_root: str
    probes: list[CandidateProbe]
    error_code: str | None = None
    message: str | None = None
    remediation: list[str] | None = None


@dataclass
class BootstrapStatus:
    """Runtime status for optional bootstrap capabilities."""

    log_enabled: bool
    state_enabled: bool
    degraded_reason: list[str]
    strict_mode: bool
    log_path: str | None = None
    state_path: str | None = None


def _is_truthy(value: str | None) -> bool:
    """Return ``True`` for environment-variable strings that represent a truthy boolean."""
    if value is None:
        return False
    return value.strip().lower() in {"1", "true", "yes", "on"}


def _is_json_mode(args: Sequence[str]) -> bool:
    """Return ``True`` when ``--json`` is present in the argument list."""
    return "--json" in args


def _resolve_optional_bootstrap_path(raw_value: str | None) -> str | None:
    """Resolve optional bootstrap file paths from environment values."""
    if raw_value is None:
        return None
    value = raw_value.strip()
    if not value:
        return None
    return os.path.abspath(os.path.expanduser(value))


def _check_optional_bootstrap_file(path: str | None, env_name: str) -> tuple[bool, str | None]:
    """Validate best-effort write capability for optional bootstrap files."""
    if path is None:
        return False, f"{env_name} is unset"
    parent = os.path.dirname(path) or "."
    try:
        os.makedirs(parent, exist_ok=True)
    except OSError as exc:
        return False, f"{env_name} parent is unavailable ({type(exc).__name__}: {exc})"
    try:
        with open(path, "a", encoding="utf8"):
            pass
    except OSError as exc:
        return False, f"{env_name} is not writable ({type(exc).__name__}: {exc})"
    return True, None


def resolve_bootstrap_status(env: dict[str, str] | None = None) -> BootstrapStatus:
    """Resolve optional bootstrap capability status from environment variables."""
    runtime_env = dict(os.environ if env is None else env)
    strict_mode = _is_truthy(runtime_env.get(BOOTSTRAP_STRICT_ENV))
    log_path = _resolve_optional_bootstrap_path(runtime_env.get("BOOTSTRAP_GLOGGUR_LOG_FILE"))
    state_path = _resolve_optional_bootstrap_path(runtime_env.get("BOOTSTRAP_GLOGGUR_STATE_FILE"))

    log_enabled, log_reason = _check_optional_bootstrap_file(log_path, "BOOTSTRAP_GLOGGUR_LOG_FILE")
    state_enabled, state_reason = _check_optional_bootstrap_file(
        state_path,
        "BOOTSTRAP_GLOGGUR_STATE_FILE",
    )
    degraded_reason = [reason for reason in (log_reason, state_reason) if reason]
    return BootstrapStatus(
        log_enabled=log_enabled,
        state_enabled=state_enabled,
        degraded_reason=degraded_reason,
        strict_mode=strict_mode,
        log_path=log_path,
        state_path=state_path,
    )


def _bootstrap_status_payload(status: BootstrapStatus) -> dict[str, object]:
    """Serialize bootstrap status into a stable diagnostics object."""
    return {
        "log_enabled": status.log_enabled,
        "state_enabled": status.state_enabled,
        "degraded_reason": list(status.degraded_reason),
        "strict_mode": status.strict_mode,
        "log_path": status.log_path,
        "state_path": status.state_path,
    }


def _error_envelope(
    *,
    error_code: str,
    error: str,
    stage: str,
    compatibility: dict[str, object] | None = None,
) -> dict[str, object]:
    """Return the normalized top-level JSON failure envelope."""
    payload: dict[str, object] = {
        "ok": False,
        "error_code": error_code,
        "error": error,
        "stage": stage,
    }
    if compatibility:
        payload["compatibility"] = compatibility
    return payload


def _repo_root() -> str:
    """Resolve the repository root from this module's on-disk location."""
    module_dir = Path(__file__).resolve().parent
    if module_dir.parent.name == "src":
        return str(module_dir.parent.parent)
    return str(module_dir.parent)


def _import_root(repo_root: str) -> str:
    """Return the directory that should be prepended to ``PYTHONPATH`` for imports.

    Returns ``<repo_root>/src`` when that directory exists (src-layout repos),
    otherwise falls back to ``repo_root`` directly.
    """
    src_root = os.path.join(repo_root, "src")
    if os.path.isdir(src_root):
        return src_root
    return repo_root


def _prepend_pythonpath(env: dict[str, str], repo_root: str) -> dict[str, str]:
    """Return a copy of ``env`` with the import root prepended to ``PYTHONPATH``.

    Deduplicates the import root so repeated calls are idempotent.
    """
    result = dict(env)
    import_root = _import_root(repo_root)
    existing = result.get("PYTHONPATH", "")
    parts = [entry for entry in existing.split(os.pathsep) if entry]
    parts = [entry for entry in parts if entry != import_root]
    result["PYTHONPATH"] = os.pathsep.join([import_root, *parts])
    return result


def _collect_required_imports(env: dict[str, str]) -> list[str]:
    """Parse the ``GLOGGUR_PREFLIGHT_REQUIRED_IMPORTS`` env var into a list of module names."""
    raw = env.get("GLOGGUR_PREFLIGHT_REQUIRED_IMPORTS", "")
    values = [item.strip() for item in raw.split(",") if item.strip()]
    return values


def _resolve_system_candidates(env: dict[str, str]) -> list[str]:
    """Return a deduplicated list of system Python interpreter paths to probe.

    Reads ``GLOGGUR_PREFLIGHT_SYSTEM_PYTHONS`` (colon-separated) when set;
    defaults to ``["python3", "python"]``.  PATH resolution is applied to
    bare names so the returned list contains absolute paths where available.
    """
    raw = env.get("GLOGGUR_PREFLIGHT_SYSTEM_PYTHONS")
    entries: list[str] = []
    if raw:
        entries = [item.strip() for item in raw.split(os.pathsep) if item.strip()]
    else:
        entries = ["python3", "python"]

    resolved: list[str] = []
    seen: set[str] = set()
    for entry in entries:
        candidate = entry
        if os.path.sep not in entry:
            resolved_candidate = shutil.which(entry)
            if resolved_candidate:
                candidate = resolved_candidate
            else:
                candidate = entry
        if candidate in seen:
            continue
        seen.add(candidate)
        resolved.append(candidate)
    return resolved


def _resolve_venv_python(repo_root: str, env: dict[str, str]) -> str:
    """Return the expected venv Python path, honoring explicit overrides."""
    override = env.get("GLOGGUR_PREFLIGHT_VENV_PYTHON")
    if override:
        return override
    return os.path.join(repo_root, ".venv", "bin", "python")


def _probe_script(module: str, required_imports: Sequence[str]) -> str:
    """Build a Python one-liner that imports ``module`` and any ``required_imports``.

    The resulting script is passed to the candidate interpreter via ``-c`` to
    verify that the runtime can successfully import gloggur and its dependencies.
    """
    commands = [
        "import importlib",
        f"importlib.import_module({module!r})",
    ]
    for item in required_imports:
        commands.append(f"importlib.import_module({item!r})")
    return "; ".join(commands)


def _probe_failure_reason(output: str) -> str:
    """Classify probe stderr/stdout into a stable failure reason code."""
    if "ModuleNotFoundError" in output or "ImportError" in output:
        return "missing_package"
    return "broken_environment"


def _probe_candidate(
    candidate_type: str,
    interpreter: str,
    module: str,
    repo_root: str,
    env: dict[str, str],
    required_imports: Sequence[str],
) -> CandidateProbe:
    """Probe a single interpreter candidate and return a ``CandidateProbe`` result.

    Runs ``interpreter -c <probe_script>`` with a 15-second timeout.  Returns a
    healthy probe on exit code 0, or a failure probe with a classified reason
    code on any error (not found, import failure, timeout, OS error).
    """
    if not interpreter:
        return CandidateProbe(
            candidate_type=candidate_type,
            interpreter=interpreter,
            module=module,
            exists=False,
            healthy=False,
            reason="interpreter_not_found",
            detail="empty interpreter path",
        )

    exists = (
        os.path.exists(interpreter)
        if os.path.sep in interpreter
        else bool(shutil.which(interpreter))
    )
    if not exists:
        return CandidateProbe(
            candidate_type=candidate_type,
            interpreter=interpreter,
            module=module,
            exists=False,
            healthy=False,
            reason="interpreter_not_found",
            detail=f"interpreter not found: {interpreter}",
        )

    probe_env = _prepend_pythonpath(env, repo_root)
    command = [interpreter, "-c", _probe_script(module, required_imports)]
    try:
        completed = subprocess.run(
            command,
            capture_output=True,
            text=True,
            cwd=repo_root,
            env=probe_env,
            timeout=15,
            check=False,
        )
    except FileNotFoundError as exc:
        return CandidateProbe(
            candidate_type=candidate_type,
            interpreter=interpreter,
            module=module,
            exists=False,
            healthy=False,
            reason="interpreter_not_found",
            detail=str(exc),
        )
    except subprocess.TimeoutExpired as exc:
        return CandidateProbe(
            candidate_type=candidate_type,
            interpreter=interpreter,
            module=module,
            exists=True,
            healthy=False,
            reason="broken_environment",
            detail=f"probe timed out: {exc}",
        )
    except OSError as exc:
        return CandidateProbe(
            candidate_type=candidate_type,
            interpreter=interpreter,
            module=module,
            exists=True,
            healthy=False,
            reason="broken_environment",
            detail=str(exc),
        )

    if completed.returncode == 0:
        return CandidateProbe(
            candidate_type=candidate_type,
            interpreter=interpreter,
            module=module,
            exists=True,
            healthy=True,
            returncode=0,
        )

    output = "\n".join(
        part.strip()
        for part in (completed.stderr or "", completed.stdout or "")
        if part and part.strip()
    )
    reason = _probe_failure_reason(output)
    detail_line = (
        output.splitlines()[-1] if output else f"probe failed with exit code {completed.returncode}"
    )
    return CandidateProbe(
        candidate_type=candidate_type,
        interpreter=interpreter,
        module=module,
        exists=True,
        healthy=False,
        reason=reason,
        detail=detail_line,
        returncode=completed.returncode,
    )


def _classify_failure(probes: Sequence[CandidateProbe], venv_exists: bool) -> str:
    """Derive the top-level failure error code from a set of failed probes.

    Priority order: missing_python > missing_package > missing_venv > broken_environment.
    """
    any_existing = any(probe.exists for probe in probes)
    if not any_existing:
        return "missing_python"

    if any(probe.reason == "missing_package" for probe in probes):
        return "missing_package"

    if not venv_exists and not any(
        probe.healthy for probe in probes if probe.candidate_type == "system"
    ):
        return "missing_venv"

    return "broken_environment"


def _failure_message(error_code: str, repo_root: str, venv_python: str) -> str:
    """Return a human-readable one-sentence failure description for the given error code."""
    if error_code == "missing_python":
        return "No usable Python interpreter was found for gloggur preflight."
    if error_code == "missing_venv":
        return (
            "Repository virtualenv is missing and no healthy fallback runtime was found "
            f"(expected {venv_python})."
        )
    if error_code == "missing_package":
        return "Python runtime found, but required gloggur dependencies are missing."
    return f"Detected a broken Python environment while preparing gloggur in {repo_root}."


def _remediation_steps(error_code: str, repo_root: str) -> list[str]:
    """Return an ordered list of remediation steps for the given error code."""
    bootstrap = os.path.join(repo_root, "scripts", "bootstrap_gloggur_env.sh")
    readiness = "python scripts/check_startup_readiness.py --format json"
    common = [
        f"Run `{bootstrap}` to create/repair .venv and install dependencies.",
        "Re-run `scripts/gloggur status --json` after bootstrap completes.",
        f"Confirm startup readiness with `{readiness}`.",
    ]
    if error_code == "missing_python":
        return [
            "Install Python 3.10+ and ensure `python3` is on PATH.",
            *common,
        ]
    if error_code == "missing_package":
        return [
            *common,
            (
                "If using a custom interpreter, install with "
                "`pip install -e '.[all,dev]'` in repo root."
            ),
        ]
    if error_code == "missing_venv":
        return common
    return [
        *common,
        "If issues persist, remove `.venv` and rerun bootstrap with `--recreate`.",
    ]


def build_launch_plan(
    args: Sequence[str],
    env: dict[str, str] | None = None,
    repo_root: str | None = None,
) -> LaunchPlan:
    """Probe available Python interpreters and return a ready-or-failed ``LaunchPlan``.

    Probes the venv interpreter first; falls back to system Python candidates
    in order.  Returns a ``LaunchPlan`` with ``ready=True`` and a selected
    interpreter as soon as a healthy candidate is found, or ``ready=False``
    with a classified error code and remediation steps if all probes fail.
    """
    runtime_env = dict(os.environ if env is None else env)
    root = repo_root or _repo_root()
    venv_python = _resolve_venv_python(root, runtime_env)
    venv_exists = os.path.exists(venv_python)
    required_imports = _collect_required_imports(runtime_env)
    probe_module = runtime_env.get("GLOGGUR_PREFLIGHT_PROBE_MODULE", PROBE_MODULE_DEFAULT)
    probes: list[CandidateProbe] = []

    venv_probe = _probe_candidate(
        candidate_type="venv",
        interpreter=venv_python,
        module=probe_module,
        repo_root=root,
        env=runtime_env,
        required_imports=required_imports,
    )
    probes.append(venv_probe)
    launch_env = _prepend_pythonpath(runtime_env, root)
    if venv_probe.healthy:
        return LaunchPlan(
            ready=True,
            args=list(args),
            interpreter=venv_python,
            module=VENV_EXEC_MODULE,
            candidate_type="venv",
            env=launch_env,
            repo_root=root,
            probes=probes,
        )

    for system_python in _resolve_system_candidates(runtime_env):
        system_probe = _probe_candidate(
            candidate_type="system",
            interpreter=system_python,
            module=probe_module,
            repo_root=root,
            env=runtime_env,
            required_imports=required_imports,
        )
        probes.append(system_probe)
        if system_probe.healthy:
            return LaunchPlan(
                ready=True,
                args=list(args),
                interpreter=system_probe.interpreter,
                module=SYSTEM_EXEC_MODULE,
                candidate_type="system",
                env=launch_env,
                repo_root=root,
                probes=probes,
            )

    error_code = _classify_failure(probes=probes, venv_exists=venv_exists)
    return LaunchPlan(
        ready=False,
        args=list(args),
        interpreter=None,
        module=None,
        candidate_type=None,
        env=launch_env,
        repo_root=root,
        probes=probes,
        error_code=error_code,
        message=_failure_message(error_code, root, venv_python),
        remediation=_remediation_steps(error_code, root),
    )


def _serialize_probes(probes: Sequence[CandidateProbe]) -> list[dict[str, object]]:
    """Convert a sequence of ``CandidateProbe`` instances to plain dicts for JSON output."""
    return [asdict(probe) for probe in probes]


def build_failure_payload(
    plan: LaunchPlan,
    *,
    preflight_ms: int,
    bootstrap_status: BootstrapStatus | None = None,
) -> dict[str, object]:
    """Build structured failure payload for a failed ``LaunchPlan``."""
    assert not plan.ready
    venv_python = _resolve_venv_python(plan.repo_root, plan.env)
    system_candidates = _resolve_system_candidates(plan.env)
    compatibility_payload: dict[str, object] = {
        "operation": CHECK_OPERATION,
        "error_code": plan.error_code,
        "message": plan.message,
        "remediation": plan.remediation or [],
        "detected_environment": {
            "repo_root": plan.repo_root,
            "venv_python": venv_python,
            "venv_exists": os.path.exists(venv_python),
            "system_candidates": system_candidates,
            "pythonpath": plan.env.get("PYTHONPATH"),
            "probes": _serialize_probes(plan.probes),
        },
        "preflight_ms": preflight_ms,
    }
    if bootstrap_status is not None:
        compatibility_payload["bootstrap_status"] = _bootstrap_status_payload(bootstrap_status)

    return _error_envelope(
        error_code=str(plan.error_code),
        error=str(plan.message),
        stage=BOOTSTRAP_STAGE,
        compatibility=compatibility_payload,
    )


def build_ready_payload(
    plan: LaunchPlan,
    *,
    preflight_ms: int,
    bootstrap_status: BootstrapStatus | None = None,
) -> dict[str, object]:
    """Build the structured success payload for a ready ``LaunchPlan``."""
    assert plan.ready
    payload: dict[str, object] = {
        "operation": CHECK_OPERATION,
        "ready": True,
        "selected_candidate": plan.candidate_type,
        "selected_interpreter": plan.interpreter,
        "selected_module": plan.module,
        "argv": plan.args,
        "preflight_ms": preflight_ms,
        "detected_environment": {
            "repo_root": plan.repo_root,
            "pythonpath": plan.env.get("PYTHONPATH"),
            "probes": _serialize_probes(plan.probes),
        },
    }
    if bootstrap_status is not None:
        payload["bootstrap_status"] = _bootstrap_status_payload(bootstrap_status)
    return payload


def build_bootstrap_degraded_payload(
    *,
    status: BootstrapStatus,
    repo_root: str,
    preflight_ms: int,
) -> dict[str, object]:
    """Build strict-mode bootstrap failure payload when optional capabilities degrade."""
    message = "Bootstrap capability degradation detected in strict mode."
    compatibility_payload = {
        "operation": CHECK_OPERATION,
        "error_code": "bootstrap_capability_degraded",
        "message": message,
        "remediation": [
            (
                "Unset BOOTSTRAP_GLOGGUR_LOG_FILE/BOOTSTRAP_GLOGGUR_STATE_FILE "
                "to disable optional writes."
            ),
            "Or set writable bootstrap paths and rerun the command.",
            "Set GLOGGUR_BOOTSTRAP_STRICT=0 to allow degraded bootstrap execution.",
        ],
        "detected_environment": {
            "repo_root": repo_root,
            "bootstrap_status": _bootstrap_status_payload(status),
        },
        "preflight_ms": preflight_ms,
    }
    return _error_envelope(
        error_code="bootstrap_capability_degraded",
        error=message,
        stage=BOOTSTRAP_STAGE,
        compatibility=compatibility_payload,
    )


def _emit(payload: dict[str, object], as_json: bool) -> None:
    """Print ``payload`` as JSON to stdout or as a human-readable summary to stderr."""
    if as_json:
        print(json.dumps(payload, separators=(",", ":"), ensure_ascii=True))
        return

    if payload.get("ready") is True:
        print(
            "gloggur preflight OK: "
            f"{payload.get('selected_candidate')} -> {payload.get('selected_interpreter')}",
            file=sys.stderr,
        )
        return

    compatibility = payload.get("compatibility")
    remediation: list[str] = []
    if isinstance(compatibility, dict):
        raw_remediation = compatibility.get("remediation", [])
        if isinstance(raw_remediation, list):
            remediation = [str(step) for step in raw_remediation]

    print(
        f"gloggur preflight failed ({payload.get('error_code')}): {payload.get('error')}",
        file=sys.stderr,
    )
    print("Remediation:", file=sys.stderr)
    for index, step in enumerate(remediation, start=1):
        print(f"{index}. {step}", file=sys.stderr)


def _execute(plan: LaunchPlan) -> None:
    """Exec into the selected interpreter via ``os.execvpe``, replacing the current process.

    This function does not return on success.
    """
    assert plan.ready
    assert plan.interpreter is not None
    assert plan.module is not None
    command = [plan.interpreter, "-m", plan.module, *plan.args]
    os.execvpe(plan.interpreter, command, plan.env)


def main(argv: Sequence[str] | None = None) -> int:
    """Entry point for the gloggur preflight bootstrap launcher."""
    args = list(sys.argv[1:] if argv is None else argv)
    as_json = _is_json_mode(args)
    started = time.perf_counter()
    status = resolve_bootstrap_status()
    if status.strict_mode and status.degraded_reason:
        preflight_ms = int((time.perf_counter() - started) * 1000)
        _emit(
            build_bootstrap_degraded_payload(
                status=status,
                repo_root=_repo_root(),
                preflight_ms=preflight_ms,
            ),
            as_json=as_json,
        )
        return 1

    try:
        plan = build_launch_plan(args=args)
    except Exception as exc:  # pragma: no cover - defensive path
        preflight_ms = int((time.perf_counter() - started) * 1000)
        payload = _error_envelope(
            error_code="broken_environment",
            error="Unexpected error while running gloggur preflight.",
            stage=BOOTSTRAP_STAGE,
            compatibility={
                "operation": CHECK_OPERATION,
                "error_code": "broken_environment",
                "message": "Unexpected error while running gloggur preflight.",
                "remediation": _remediation_steps("broken_environment", _repo_root()),
                "detected_environment": {
                    "repo_root": _repo_root(),
                    "exception": repr(exc),
                    "bootstrap_status": _bootstrap_status_payload(status),
                },
                "preflight_ms": preflight_ms,
            },
        )
        _emit(payload, as_json=as_json)
        return EXIT_CODES["broken_environment"]

    preflight_ms = int((time.perf_counter() - started) * 1000)
    dry_run = _is_truthy(os.environ.get("GLOGGUR_PREFLIGHT_DRY_RUN"))
    if plan.ready:
        if dry_run:
            _emit(
                build_ready_payload(plan, preflight_ms=preflight_ms, bootstrap_status=status),
                as_json=as_json,
            )
            return 0
        _execute(plan)
        return 0

    payload = build_failure_payload(
        plan,
        preflight_ms=preflight_ms,
        bootstrap_status=status,
    )
    _emit(payload, as_json=as_json)
    return EXIT_CODES.get(str(plan.error_code), EXIT_CODES["broken_environment"])


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

from gloggur.search.compat import attach_legacy_search_contract
from gloggur.search.router import (
    ContextPack,
    SearchConstraints,
    SearchIntent,
    SearchRouter,
    SearchRouterConfig,
    load_search_router_config,
)

__all__ = [
    "attach_legacy_search_contract",
    "ContextPack",
    "SearchConstraints",
    "SearchIntent",
    "SearchRouter",
    "SearchRouterConfig",
    "load_search_router_config",
]



from __future__ import annotations

import json
import os
import re
import shutil
import subprocess
import sys
import tarfile
import tempfile
import threading
import time
from collections.abc import Callable, Sequence
from dataclasses import asdict, dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

from gloggur import __version__ as GLOGGUR_VERSION
from gloggur.archive_utils import (
    ArchiveFileSource,
    create_deterministic_tar_gz,
    sha256_bytes,
    sha256_file,
)
from gloggur.bootstrap_launcher import resolve_bootstrap_status
from gloggur.byte_spans import discover_repo_root
from gloggur.config import GloggurConfig
from gloggur.indexer.concurrency import read_cache_lock_metadata
from gloggur.io_failures import StorageIOError, wrap_io_error
from gloggur.search.router.config import load_search_router_config
from gloggur.support_runtime import (
    capture_enabled as runtime_capture_enabled,
)
from gloggur.support_runtime import (
    load_runtime_records,
    request_active_stack_dumps,
    should_auto_include_cache,
    support_runtime_paths,
)

SUPPORT_MANIFEST_SCHEMA_VERSION = "1"
SUPPORT_BUNDLE_TYPE = "support"
DEFAULT_TAIL_BYTES = 5 * 1024 * 1024
_SESSION_ID_RE = re.compile(r"^[A-Za-z0-9._-]+$")
_SECRET_KEY_RE = re.compile(r"(?:^|_)(?:key|token|secret|password)$", re.IGNORECASE)
_REPO_ROOT_PLACEHOLDER = "<REPO_ROOT>"
_HOME_PLACEHOLDER = "<HOME>"
_REDACTED_PLACEHOLDER = "<REDACTED>"
_CAPTURE_META_PATH = "diagnostics/capture_meta.json"
_ALLOWLISTED_ENV_KEYS = (
    "GLOGGUR_CACHE_DIR",
    "GLOGGUR_EMBEDDING_PROVIDER",
    "GLOGGUR_WATCH_ENABLED",
    "GLOGGUR_WATCH_PATH",
    "GLOGGUR_WATCH_MODE",
    "GLOGGUR_WATCH_STATE_FILE",
    "GLOGGUR_WATCH_LOG_FILE",
    "GLOGGUR_STORAGE_BACKEND",
    "GLOGGUR_RUNTIME_HOST",
    "GLOGGUR_RUN_FROM_CALLER_CWD",
    "BOOTSTRAP_GLOGGUR_LOG_FILE",
    "BOOTSTRAP_GLOGGUR_STATE_FILE",
)


class SupportContractError(RuntimeError):
    """Stable contract failure for support session commands."""

    def __init__(self, message: str, *, code: str) -> None:
        super().__init__(message)
        self.code = code


@dataclass(frozen=True)
class SupportRoots:
    """Filesystem roots used by support sessions and bundles."""

    repo_root: Path
    support_root: Path
    sessions_root: Path
    bundles_root: Path
    runtime_root: Path
    active_root: Path
    recent_root: Path


@dataclass(frozen=True)
class SupportCallbacks:
    """Late-bound callbacks supplied by the CLI entrypoints."""

    load_config: Callable[[str | None], tuple[GloggurConfig, str | None]]
    build_status_payload: Callable[[GloggurConfig], dict[str, object]]
    build_watch_status_payload: Callable[[GloggurConfig], dict[str, object]]


def run_support_command(
    *,
    as_json: bool,
    child_args: Sequence[str],
    note: str | None,
    bundle_on_failure: bool,
    destination: str | None,
    callbacks: SupportCallbacks,
) -> tuple[dict[str, object], int]:
    """Run a traced Glöggur child command inside a support session."""
    normalized_child_args = _validate_child_args(child_args)
    config_path = _extract_config_path(normalized_child_args)
    roots = _support_roots(discover_repo_root())
    session_dir = _create_session_dir(roots)
    session_id = session_dir.name
    cwd = Path.cwd().resolve()
    session_payload = {
        "session_id": session_id,
        "mode": "run",
        "repo_root": _sanitize_path(str(roots.repo_root), roots.repo_root),
        "cwd": _sanitize_path(str(cwd), roots.repo_root),
        "child_argv": list(normalized_child_args),
        "child_command": normalized_child_args[0],
        "config_path": _sanitize_optional_string(config_path, roots.repo_root),
        "created_at": _utc_now_iso(),
        "status": "running",
        "note_present": bool(note),
    }
    _write_json(session_dir / "session.json", session_payload)
    if note:
        _append_note(session_dir / "notes.txt", note, roots.repo_root)

    bootstrap_log = session_dir / "logs" / "bootstrap.log"
    bootstrap_state = session_dir / "logs" / "bootstrap.state.json"
    command = _resolve_child_command(roots.repo_root, normalized_child_args)
    env = os.environ.copy()
    env["BOOTSTRAP_GLOGGUR_LOG_FILE"] = str(bootstrap_log)
    env["BOOTSTRAP_GLOGGUR_STATE_FILE"] = str(bootstrap_state)
    env["GLOGGUR_RUN_FROM_CALLER_CWD"] = "1"

    started = time.perf_counter()
    if as_json:
        completed = _run_child_captured(command, cwd=cwd, env=env)
        raw_stdout = completed.stdout
        raw_stderr = completed.stderr
        exit_code = int(completed.returncode)
        _write_sanitized_text_log(
            session_dir / "logs" / "child.stdout.log",
            raw_stdout,
            repo_root=roots.repo_root,
            session_dir=session_dir,
        )
        _write_sanitized_text_log(
            session_dir / "logs" / "child.stderr.log",
            raw_stderr,
            repo_root=roots.repo_root,
            session_dir=session_dir,
        )
    else:
        raw_stdout, raw_stderr, exit_code = _run_child_streaming(
            command,
            cwd=cwd,
            env=env,
            stdout_log_path=session_dir / "logs" / "child.stdout.log",
            stderr_log_path=session_dir / "logs" / "child.stderr.log",
            repo_root=roots.repo_root,
        )
    duration_ms = int((time.perf_counter() - started) * 1000)

    child_payload = _parse_first_json_object(raw_stdout)
    if child_payload is not None:
        _write_json(
            session_dir / "diagnostics" / "child_payload.json",
            _sanitize_object(child_payload, roots.repo_root),
        )
    child_error_code, child_failure_codes = _extract_failure_contract(child_payload)

    _capture_support_diagnostics(
        session_dir=session_dir,
        repo_root=roots.repo_root,
        callbacks=callbacks,
        config_path=config_path,
        support_env=env,
        note=None,
    )
    _snapshot_optional_runtime_files(
        session_dir=session_dir,
        repo_root=roots.repo_root,
        config_path=config_path,
        callbacks=callbacks,
    )

    bundle_payload: dict[str, object] | None = None
    bundle_error: dict[str, object] | None = None
    if exit_code != 0 and bundle_on_failure:
        try:
            bundle_payload = collect_support_bundle(
                session_id=session_id,
                note=None,
                destination=destination,
                include_cache=False,
                overwrite=False,
                callbacks=callbacks,
                config_path=config_path,
            )
        except (SupportContractError, StorageIOError) as exc:
            bundle_error = _bundle_error_payload(exc)

    session_payload.update(
        {
            "status": "completed" if exit_code == 0 else "failed",
            "completed_at": _utc_now_iso(),
            "duration_ms": duration_ms,
            "exit_code": exit_code,
            "child_error_code": child_error_code,
            "child_failure_codes": child_failure_codes,
            "bundle_created": bundle_payload is not None,
            "bundle_path": bundle_payload.get("bundle_path") if bundle_payload else None,
            "bundle_error": bundle_error,
        }
    )
    _write_json(session_dir / "session.json", session_payload)

    result = {
        "session_id": session_id,
        "session_dir": str(session_dir),
        "child_argv": list(normalized_child_args),
        "child_exit_code": exit_code,
        "child_error_code": child_error_code,
        "failure_codes": child_failure_codes,
        "bundle_created": bundle_payload is not None,
        "bundle_path": bundle_payload.get("bundle_path") if bundle_payload else None,
        "bundle_manifest_sha256": bundle_payload.get("manifest_sha256") if bundle_payload else None,
        "bundle_error": bundle_error,
    }
    return result, exit_code


def collect_support_bundle(
    *,
    session_id: str | None,
    note: str | None,
    destination: str | None,
    include_cache: bool,
    allow_sensitive_data: bool = False,
    overwrite: bool,
    callbacks: SupportCallbacks,
    config_path: str | None = None,
) -> dict[str, object]:
    """Collect diagnostics for a session and package a deterministic support bundle."""
    roots = _support_roots(discover_repo_root())
    support_mode_enabled = runtime_capture_enabled(roots.repo_root)
    if session_id is None:
        session_dir = _create_session_dir(roots)
        session_payload = {
            "session_id": session_dir.name,
            "mode": "manual_snapshot",
            "repo_root": _sanitize_path(str(roots.repo_root), roots.repo_root),
            "cwd": _sanitize_path(str(Path.cwd().resolve()), roots.repo_root),
            "created_at": _utc_now_iso(),
            "status": "collecting",
            "child_argv": [],
            "child_command": None,
            "config_path": _sanitize_optional_string(config_path, roots.repo_root),
            "note_present": bool(note),
        }
        _write_json(session_dir / "session.json", session_payload)
    else:
        session_dir = _resolve_existing_session_dir(roots, session_id)
        session_payload = _read_session_payload(session_dir)
    if note:
        _append_note(session_dir / "notes.txt", note, roots.repo_root)
        session_payload["note_present"] = True

    _capture_support_diagnostics(
        session_dir=session_dir,
        repo_root=roots.repo_root,
        callbacks=callbacks,
        config_path=config_path,
        support_env=None,
        note=None,
    )
    _snapshot_optional_runtime_files(
        session_dir=session_dir,
        repo_root=roots.repo_root,
        config_path=config_path,
        callbacks=callbacks,
    )
    _capture_cache_lock_diagnostics(
        session_dir=session_dir,
        repo_root=roots.repo_root,
        callbacks=callbacks,
        config_path=config_path,
    )

    active_records, recent_records = load_runtime_records(roots.repo_root)
    capture_warnings: list[str] = []
    if support_mode_enabled:
        capture_warnings.extend(request_active_stack_dumps(roots.repo_root))
        active_records, recent_records = load_runtime_records(roots.repo_root)
    else:
        capture_warnings.append("betatester_support_not_enabled")
    capture_warnings.extend(
        _snapshot_runtime_trace_dirs(
            roots=roots,
            session_dir=session_dir,
        )
    )

    status_payload = _extract_status_payload(session_dir)
    auto_include_cache = False
    if not include_cache:
        auto_include_cache = should_auto_include_cache(
            active_records=active_records,
            recent_records=recent_records,
            status_payload=status_payload,
        )
    effective_include_cache = include_cache or auto_include_cache
    bundle_policy_applied = "explicit"
    if not include_cache:
        bundle_policy_applied = "smart" if effective_include_cache else "sanitized"

    bundle_path = _resolve_bundle_destination(
        roots=roots,
        session_id=session_dir.name,
        destination=destination,
        overwrite=overwrite,
    )
    config: GloggurConfig | None = None
    try:
        config, _ = callbacks.load_config(config_path)
    except (StorageIOError, OSError, ValueError, TypeError):
        config = None
    cache_dir = config.cache_dir if config is not None else str(roots.repo_root / ".gloggur-cache")
    runtime_artifacts = _inventory_runtime_artifacts(roots.repo_root, cache_dir)
    included_artifacts = runtime_artifacts if effective_include_cache else []
    excluded_artifacts = [] if effective_include_cache else runtime_artifacts
    _write_bundle(
        repo_root=roots.repo_root,
        session_dir=session_dir,
        bundle_path=bundle_path,
        include_cache=effective_include_cache,
        allow_sensitive_data=allow_sensitive_data,
        included_artifacts=included_artifacts,
        excluded_artifacts=excluded_artifacts,
        support_mode_enabled=support_mode_enabled,
        bundle_policy_applied=bundle_policy_applied,
        active_commands=_summarize_command_records(active_records),
        capture_warnings=capture_warnings,
    )
    manifest_payload = _read_bundle_manifest(bundle_path)
    public_included_artifacts = _public_artifact_inventory(included_artifacts)
    public_excluded_artifacts = _public_artifact_inventory(excluded_artifacts)

    session_payload.update(
        {
            "status": "collected",
            "collected_at": _utc_now_iso(),
            "bundle_path": str(bundle_path),
            "redaction_mode": "sanitized",
            "include_cache": effective_include_cache,
            "bundle_sensitivity": "sensitive" if effective_include_cache else "sanitized",
            "includes_cache_artifacts": effective_include_cache,
            "sensitive_data_acknowledged": allow_sensitive_data,
            "excluded_artifacts": public_excluded_artifacts,
            "included_artifacts": public_included_artifacts,
            "support_mode_enabled": support_mode_enabled,
            "bundle_policy_applied": bundle_policy_applied,
            "active_commands": _summarize_command_records(active_records),
            "capture_warnings": capture_warnings,
        }
    )
    _write_json(session_dir / "session.json", session_payload)

    return {
        "collected": True,
        "session_id": session_dir.name,
        "session_dir": str(session_dir),
        "bundle_path": str(bundle_path),
        "bundle_uri": bundle_path.resolve().as_uri(),
        "include_cache": effective_include_cache,
        "bundle_sensitivity": "sensitive" if effective_include_cache else "sanitized",
        "includes_cache_artifacts": effective_include_cache,
        "sensitive_data_acknowledged": allow_sensitive_data,
        "manifest_sha256": manifest_payload["manifest_sha256"],
        "archive_sha256": manifest_payload["archive_sha256"],
        "archive_bytes": manifest_payload["archive_bytes"],
        "excluded_artifacts": public_excluded_artifacts,
        "included_artifacts": public_included_artifacts,
        "support_mode_enabled": support_mode_enabled,
        "bundle_policy_applied": bundle_policy_applied,
        "active_commands": _summarize_command_records(active_records),
        "capture_warnings": capture_warnings,
    }


def _support_roots(repo_root: Path) -> SupportRoots:
    support_root = repo_root / ".gloggur" / "support"
    runtime_paths = support_runtime_paths(repo_root)
    return SupportRoots(
        repo_root=repo_root,
        support_root=support_root,
        sessions_root=support_root / "sessions",
        bundles_root=support_root / "bundles",
        runtime_root=runtime_paths.runtime_root,
        active_root=runtime_paths.active_root,
        recent_root=runtime_paths.recent_root,
    )


def _validate_child_args(child_args: Sequence[str]) -> list[str]:
    normalized = [str(arg) for arg in child_args if str(arg)]
    if not normalized:
        raise SupportContractError(
            "Support run requires a Glöggur child command after `--`.",
            code="support_command_invalid",
        )
    first = normalized[0]
    if first in {"gloggur", "support"}:
        raise SupportContractError(
            "Support run only accepts Glöggur subcommand argv.",
            code="support_command_invalid",
        )
    if first.startswith("-") or "/" in first or "\\" in first:
        raise SupportContractError(
            "Support run only accepts Glöggur subcommand argv, not executable paths.",
            code="support_command_invalid",
        )
    return normalized


def _extract_config_path(child_args: Sequence[str]) -> str | None:
    for index, arg in enumerate(child_args):
        if arg == "--config" and index + 1 < len(child_args):
            return child_args[index + 1]
        if arg.startswith("--config="):
            return arg.split("=", 1)[1]
    return None


def _resolve_child_command(repo_root: Path, child_args: Sequence[str]) -> list[str]:
    wrapper = repo_root / "scripts" / "gloggur"
    if wrapper.exists():
        return [str(wrapper), *child_args]
    return [sys.executable, "-m", "gloggur.cli.main", *child_args]


def _create_session_dir(roots: SupportRoots) -> Path:
    base_id = f"{datetime.now(timezone.utc).strftime('%Y%m%dT%H%M%SZ')}-{os.getpid()}"
    session_id = base_id
    candidate = roots.sessions_root / session_id
    suffix = 1
    while candidate.exists():
        session_id = f"{base_id}-{suffix}"
        candidate = roots.sessions_root / session_id
        suffix += 1
    for path in (candidate, candidate / "logs", candidate / "diagnostics"):
        try:
            path.mkdir(parents=True, exist_ok=True)
        except OSError as exc:
            raise wrap_io_error(
                exc, operation="create support session directory", path=str(path)
            ) from exc
    return candidate


def _resolve_existing_session_dir(roots: SupportRoots, session_id: str) -> Path:
    normalized = session_id.strip()
    if not normalized or not _SESSION_ID_RE.fullmatch(normalized):
        raise SupportContractError(
            f"Support session id is invalid: {session_id!r}",
            code="support_session_invalid",
        )
    session_dir = (roots.sessions_root / normalized).resolve()
    try:
        within_root = os.path.commonpath(
            [str(roots.sessions_root.resolve()), str(session_dir)]
        ) == str(roots.sessions_root.resolve())
    except ValueError as exc:
        raise SupportContractError(
            f"Support session id escapes the support root: {session_id!r}",
            code="support_session_invalid",
        ) from exc
    if not within_root:
        raise SupportContractError(
            f"Support session id escapes the support root: {session_id!r}",
            code="support_session_invalid",
        )
    if not session_dir.exists():
        raise SupportContractError(
            f"Support session does not exist: {normalized}",
            code="support_session_missing",
        )
    if not (session_dir / "session.json").exists():
        raise SupportContractError(
            f"Support session is missing session.json: {normalized}",
            code="support_session_invalid",
        )
    return session_dir


def _capture_support_diagnostics(
    *,
    session_dir: Path,
    repo_root: Path,
    callbacks: SupportCallbacks,
    config_path: str | None,
    support_env: dict[str, str] | None,
    note: str | None,
) -> None:
    if note:
        _append_note(session_dir / "notes.txt", note, repo_root)

    config_payload: dict[str, object]
    config: GloggurConfig | None = None
    resolved_config_path: str | None = None
    try:
        config, resolved_config_path = callbacks.load_config(config_path)
        config_payload = {
            "captured": True,
            "config_path": _sanitize_optional_string(resolved_config_path, repo_root),
            "config": _sanitize_object(asdict(config), repo_root),
            "search_router": _sanitize_object(
                asdict(load_search_router_config(repo_root)), repo_root
            ),
            "env_summary": _sanitize_object(_collect_env_summary(support_env), repo_root),
        }
    except (StorageIOError, OSError, ValueError, TypeError) as exc:
        config_payload = _diagnostic_error_payload(exc, repo_root)
    _write_json(session_dir / "diagnostics" / "config.redacted.json", config_payload)

    bootstrap_payload = {
        "captured": True,
        "bootstrap_status": _sanitize_object(
            asdict(resolve_bootstrap_status(env=support_env or os.environ)),
            repo_root,
        ),
    }
    _write_json(session_dir / "diagnostics" / "bootstrap_status.json", bootstrap_payload)

    status_payload = _capture_callback_payload(
        callback=((lambda: callbacks.build_status_payload(config)) if config is not None else None),
        repo_root=repo_root,
    )
    _write_json(session_dir / "diagnostics" / "status.json", status_payload)

    watch_payload = _capture_callback_payload(
        callback=(
            (lambda: callbacks.build_watch_status_payload(config)) if config is not None else None
        ),
        repo_root=repo_root,
    )
    _write_json(session_dir / "diagnostics" / "watch_status.json", watch_payload)


def _capture_callback_payload(
    *,
    callback: Callable[[], dict[str, object]] | None,
    repo_root: Path,
) -> dict[str, object]:
    if callback is None:
        return {"captured": False, "reason": "config_unavailable"}
    try:
        return {"captured": True, "payload": _sanitize_object(callback(), repo_root)}
    except (StorageIOError, OSError, ValueError, TypeError) as exc:
        return _diagnostic_error_payload(exc, repo_root)


def _snapshot_optional_runtime_files(
    *,
    session_dir: Path,
    repo_root: Path,
    config_path: str | None,
    callbacks: SupportCallbacks,
) -> None:
    config: GloggurConfig | None = None
    try:
        config, _ = callbacks.load_config(config_path)
    except (StorageIOError, OSError, ValueError, TypeError):
        config = None
    router_config = load_search_router_config(repo_root)
    log_sources: list[tuple[Path, Path, bool]] = []
    if config is not None:
        log_sources.extend(
            [
                (Path(config.watch_log_file), session_dir / "logs" / "watch.log", True),
                (Path(config.watch_state_file), session_dir / "logs" / "watch_state.json", True),
                (Path(config.watch_pid_file), session_dir / "logs" / "watch.pid", True),
            ]
        )
    router_log_path = repo_root / router_config.log_path
    log_sources.append((router_log_path, session_dir / "logs" / "search_router.jsonl", True))
    bootstrap_log_path = os.environ.get("BOOTSTRAP_GLOGGUR_LOG_FILE")
    if bootstrap_log_path:
        log_sources.append((Path(bootstrap_log_path), session_dir / "logs" / "bootstrap.log", True))
    bootstrap_state_path = os.environ.get("BOOTSTRAP_GLOGGUR_STATE_FILE")
    if bootstrap_state_path:
        log_sources.append(
            (Path(bootstrap_state_path), session_dir / "logs" / "bootstrap.state.json", True)
        )

    for source_path, destination_path, sanitize in log_sources:
        if destination_path.exists():
            continue
        if sanitize:
            _snapshot_optional_text_file(
                source_path=source_path,
                destination_path=destination_path,
                repo_root=repo_root,
                session_dir=session_dir,
            )
        else:
            _snapshot_optional_file(source_path=source_path, destination_path=destination_path)


def _capture_cache_lock_diagnostics(
    *,
    session_dir: Path,
    repo_root: Path,
    callbacks: SupportCallbacks,
    config_path: str | None,
) -> None:
    config: GloggurConfig | None = None
    try:
        config, _ = callbacks.load_config(config_path)
    except (StorageIOError, OSError, ValueError, TypeError):
        config = None
    if config is None:
        _write_json(
            session_dir / "diagnostics" / "cache_lock.json",
            {"captured": False, "reason": "config_unavailable"},
        )
        return
    lock_metadata = read_cache_lock_metadata(config.cache_dir)
    payload: dict[str, object] = {
        "captured": True,
        "lock_metadata": _sanitize_object(lock_metadata, repo_root),
    }
    _write_json(session_dir / "diagnostics" / "cache_lock.json", payload)


def _snapshot_runtime_trace_dirs(*, roots: SupportRoots, session_dir: Path) -> list[str]:
    warnings: list[str] = []
    for source_root, label in (
        (roots.active_root, "active"),
        (roots.recent_root, "recent"),
    ):
        destination_root = session_dir / "runtime" / label
        destination_root.parent.mkdir(parents=True, exist_ok=True)
        if not source_root.exists():
            continue
        for source_dir in sorted(path for path in source_root.iterdir() if path.is_dir()):
            destination_dir = destination_root / source_dir.name
            try:
                if destination_dir.exists():
                    shutil.rmtree(destination_dir)
                shutil.copytree(source_dir, destination_dir)
            except OSError:
                warnings.append(f"runtime_trace_copy_failed:{label}:{source_dir.name}")
    return warnings


def _extract_status_payload(session_dir: Path) -> dict[str, object] | None:
    status_path = session_dir / "diagnostics" / "status.json"
    if not status_path.exists():
        return None
    try:
        payload = json.loads(status_path.read_text(encoding="utf8"))
    except (OSError, json.JSONDecodeError):
        return None
    if not isinstance(payload, dict):
        return None
    inner = payload.get("payload")
    return inner if isinstance(inner, dict) else None


def _summarize_command_records(records: list[dict[str, object]]) -> list[dict[str, object]]:
    summary: list[dict[str, object]] = []
    for record in records:
        summary.append(
            {
                "command_id": record.get("command_id"),
                "command_name": record.get("command_name"),
                "pid": record.get("pid"),
                "status": record.get("status"),
                "stage": record.get("stage"),
                "updated_at": record.get("updated_at"),
                "exit_code": record.get("exit_code"),
            }
        )
    return summary


def _snapshot_optional_text_file(
    *,
    source_path: Path,
    destination_path: Path,
    repo_root: Path,
    session_dir: Path,
) -> None:
    if not source_path.exists() or not source_path.is_file():
        return
    try:
        raw_bytes = source_path.read_bytes()
    except OSError as exc:
        raise wrap_io_error(
            exc, operation="read support log source", path=str(source_path)
        ) from exc
    truncated = len(raw_bytes) > DEFAULT_TAIL_BYTES
    payload = raw_bytes[-DEFAULT_TAIL_BYTES:] if truncated else raw_bytes
    text = payload.decode("utf8", errors="replace")
    sanitized = _sanitize_text(text, repo_root)
    try:
        destination_path.parent.mkdir(parents=True, exist_ok=True)
        destination_path.write_text(sanitized, encoding="utf8")
    except OSError as exc:
        raise wrap_io_error(
            exc, operation="write support log snapshot", path=str(destination_path)
        ) from exc
    _record_capture_metadata(
        session_dir=session_dir,
        session_rel_path=destination_path.relative_to(session_dir).as_posix(),
        source_path=source_path,
        source_bytes=len(raw_bytes),
        archived_bytes=len(sanitized.encode("utf8")),
        truncated=truncated,
    )


def _snapshot_optional_file(*, source_path: Path, destination_path: Path) -> None:
    if not source_path.exists() or not source_path.is_file():
        return
    try:
        destination_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(source_path, destination_path)
    except OSError as exc:
        raise wrap_io_error(
            exc, operation="copy support runtime file", path=str(destination_path)
        ) from exc


def _collect_env_summary(env: dict[str, str] | None) -> dict[str, str]:
    source = dict(os.environ if env is None else env)
    summary: dict[str, str] = {}
    for key in _ALLOWLISTED_ENV_KEYS:
        value = source.get(key)
        if value is None or value == "":
            continue
        summary[key] = value
    return summary


def _write_sanitized_text_log(
    log_path: Path,
    raw_text: str,
    *,
    repo_root: Path,
    session_dir: Path,
) -> None:
    payload = raw_text.encode("utf8", errors="replace")
    truncated = len(payload) > DEFAULT_TAIL_BYTES
    truncated_payload = payload[-DEFAULT_TAIL_BYTES:] if truncated else payload
    sanitized = _sanitize_text(truncated_payload.decode("utf8", errors="replace"), repo_root)
    try:
        log_path.parent.mkdir(parents=True, exist_ok=True)
        log_path.write_text(sanitized, encoding="utf8")
    except OSError as exc:
        raise wrap_io_error(exc, operation="write support child log", path=str(log_path)) from exc
    _record_capture_metadata(
        session_dir=session_dir,
        session_rel_path=log_path.relative_to(session_dir).as_posix(),
        source_path=None,
        source_bytes=len(payload),
        archived_bytes=len(sanitized.encode("utf8")),
        truncated=truncated,
    )


def _run_child_captured(
    command: Sequence[str],
    *,
    cwd: Path,
    env: dict[str, str],
) -> subprocess.CompletedProcess[str]:
    try:
        return subprocess.run(
            list(command),
            cwd=str(cwd),
            env=env,
            capture_output=True,
            text=True,
            errors="replace",
            check=False,
        )
    except OSError as exc:
        raise wrap_io_error(
            exc, operation="spawn support child command", path=str(command[0])
        ) from exc


def _run_child_streaming(
    command: Sequence[str],
    *,
    cwd: Path,
    env: dict[str, str],
    stdout_log_path: Path,
    stderr_log_path: Path,
    repo_root: Path,
) -> tuple[str, str, int]:
    try:
        process = subprocess.Popen(
            list(command),
            cwd=str(cwd),
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            errors="replace",
            bufsize=1,
        )
    except OSError as exc:
        raise wrap_io_error(
            exc, operation="spawn support child command", path=str(command[0])
        ) from exc

    stdout_parts: list[str] = []
    stderr_parts: list[str] = []
    stdout_log_path.parent.mkdir(parents=True, exist_ok=True)
    stderr_log_path.parent.mkdir(parents=True, exist_ok=True)
    with (
        stdout_log_path.open("w", encoding="utf8") as stdout_log,
        stderr_log_path.open(
            "w",
            encoding="utf8",
        ) as stderr_log,
    ):
        threads = [
            threading.Thread(
                target=_relay_stream,
                args=(process.stdout, sys.stdout, stdout_parts, stdout_log, repo_root),
                daemon=True,
            ),
            threading.Thread(
                target=_relay_stream,
                args=(process.stderr, sys.stderr, stderr_parts, stderr_log, repo_root),
                daemon=True,
            ),
        ]
        for thread in threads:
            thread.start()
        exit_code = int(process.wait())
        for thread in threads:
            thread.join()
    raw_stdout = "".join(stdout_parts)
    raw_stderr = "".join(stderr_parts)
    _record_capture_metadata(
        session_dir=stdout_log_path.parent.parent,
        session_rel_path=stdout_log_path.relative_to(stdout_log_path.parent.parent).as_posix(),
        source_path=None,
        source_bytes=len(raw_stdout.encode("utf8", errors="replace")),
        archived_bytes=stdout_log_path.stat().st_size,
        truncated=False,
    )
    _record_capture_metadata(
        session_dir=stderr_log_path.parent.parent,
        session_rel_path=stderr_log_path.relative_to(stderr_log_path.parent.parent).as_posix(),
        source_path=None,
        source_bytes=len(raw_stderr.encode("utf8", errors="replace")),
        archived_bytes=stderr_log_path.stat().st_size,
        truncated=False,
    )
    return raw_stdout, raw_stderr, exit_code


def _relay_stream(
    stream: Any,
    terminal: Any,
    buffer: list[str],
    log_handle: Any,
    repo_root: Path,
) -> None:
    if stream is None:
        return
    try:
        for chunk in iter(stream.readline, ""):
            if not chunk:
                break
            buffer.append(chunk)
            terminal.write(chunk)
            terminal.flush()
            log_handle.write(_sanitize_text(chunk, repo_root))
            log_handle.flush()
        remainder = stream.read()
        if remainder:
            buffer.append(remainder)
            terminal.write(remainder)
            terminal.flush()
            log_handle.write(_sanitize_text(remainder, repo_root))
            log_handle.flush()
    finally:
        try:
            stream.close()
        except Exception:
            pass


def _write_bundle(
    *,
    repo_root: Path,
    session_dir: Path,
    bundle_path: Path,
    include_cache: bool,
    allow_sensitive_data: bool,
    included_artifacts: list[dict[str, object]],
    excluded_artifacts: list[dict[str, object]],
    support_mode_enabled: bool,
    bundle_policy_applied: str,
    active_commands: list[dict[str, object]],
    capture_warnings: list[str],
) -> None:
    support_prefix = f"support/{session_dir.name}"
    try:
        staging_dir = Path(
            tempfile.mkdtemp(
                prefix=f"support-bundle-{session_dir.name}-",
                dir=str(bundle_path.parent),
            )
        )
    except OSError as exc:
        raise wrap_io_error(
            exc,
            operation="create support bundle staging directory",
            path=str(bundle_path.parent),
        ) from exc
    try:
        shutil.copytree(session_dir, staging_dir / support_prefix)
        extra_sources: list[ArchiveFileSource] = []
        if include_cache:
            for item in included_artifacts:
                rel_path = str(item["path"])
                source_path = Path(str(item["source_path"]))
                extra_sources.append(
                    ArchiveFileSource(
                        source_path=source_path,
                        archive_path=f"{support_prefix}/artifacts/{rel_path}",
                    )
                )
        staged_sources = [
            ArchiveFileSource(
                source_path=path,
                archive_path=(
                    f"{support_prefix}/"
                    f"{path.relative_to(staging_dir / support_prefix).as_posix()}"
                ),
            )
            for path in sorted((staging_dir / support_prefix).rglob("*"))
            if path.is_file()
        ]
        all_sources = [*staged_sources, *extra_sources]
        manifest = _build_bundle_manifest(
            session_dir=session_dir,
            archive_sources=all_sources,
            include_cache=include_cache,
            allow_sensitive_data=allow_sensitive_data,
            included_artifacts=included_artifacts,
            excluded_artifacts=excluded_artifacts,
            repo_root=repo_root,
            support_mode_enabled=support_mode_enabled,
            bundle_policy_applied=bundle_policy_applied,
            active_commands=active_commands,
            capture_warnings=capture_warnings,
        )
        manifest_bytes = json.dumps(manifest, sort_keys=True, indent=2).encode("utf8") + b"\n"
        create_deterministic_tar_gz(
            bundle_path,
            file_sources=all_sources,
            extra_files=(("manifest.json", manifest_bytes),),
        )
    except OSError as exc:
        raise wrap_io_error(
            exc, operation="create support bundle archive", path=str(bundle_path)
        ) from exc
    finally:
        shutil.rmtree(staging_dir, ignore_errors=True)


def _build_bundle_manifest(
    *,
    session_dir: Path,
    archive_sources: Sequence[ArchiveFileSource],
    include_cache: bool,
    allow_sensitive_data: bool,
    included_artifacts: list[dict[str, object]],
    excluded_artifacts: list[dict[str, object]],
    repo_root: Path,
    support_mode_enabled: bool,
    bundle_policy_applied: str,
    active_commands: list[dict[str, object]],
    capture_warnings: list[str],
) -> dict[str, object]:
    capture_meta = _load_capture_metadata(session_dir)
    file_entries: list[dict[str, object]] = []
    archive_prefix = f"support/{session_dir.name}/"
    for source in sorted(archive_sources, key=lambda item: item.archive_path):
        try:
            size_bytes = source.source_path.stat().st_size
        except OSError as exc:
            raise wrap_io_error(
                exc,
                operation="read support bundle file metadata",
                path=str(source.source_path),
            ) from exc
        session_rel_path = None
        if source.archive_path.startswith(archive_prefix):
            session_rel_path = source.archive_path[len(archive_prefix) :]
        meta = capture_meta.get(session_rel_path or "", {})
        file_entries.append(
            {
                "path": source.archive_path,
                "bytes": size_bytes,
                "sha256": _sha256_file(source.source_path),
                "truncated": bool(meta.get("truncated", False)),
                "source_bytes": meta.get("source_bytes", size_bytes),
                "archived_bytes": meta.get("archived_bytes", size_bytes),
            }
        )
    session_payload = _read_session_payload(session_dir)
    manifest = {
        "manifest_schema_version": SUPPORT_MANIFEST_SCHEMA_VERSION,
        "bundle_type": SUPPORT_BUNDLE_TYPE,
        "created_at": _utc_now_iso(),
        "tool_version": GLOGGUR_VERSION,
        "session": _sanitize_object(session_payload, repo_root),
        "redaction_mode": "sanitized",
        "include_cache": include_cache,
        "support_mode_enabled": support_mode_enabled,
        "bundle_policy_applied": bundle_policy_applied,
        "bundle_sensitivity": "sensitive" if include_cache else "sanitized",
        "includes_cache_artifacts": include_cache,
        "sensitive_data_acknowledged": allow_sensitive_data,
        "included_artifacts": _public_artifact_inventory(included_artifacts),
        "excluded_artifacts": _public_artifact_inventory(excluded_artifacts),
        "active_commands": active_commands,
        "capture_warnings": capture_warnings,
        "files_total": len(file_entries),
        "bytes_total": sum(int(entry["bytes"]) for entry in file_entries),
        "files": file_entries,
    }
    return manifest


def _resolve_bundle_destination(
    *,
    roots: SupportRoots,
    session_id: str,
    destination: str | None,
    overwrite: bool,
) -> Path:
    default_path = roots.bundles_root / f"gloggur-support-{session_id}.tar.gz"
    bundle_path = Path(destination).expanduser() if destination else default_path
    if destination:
        bundle_path = bundle_path.resolve() if bundle_path.exists() else bundle_path.absolute()
        if str(destination).endswith(os.sep) or bundle_path.is_dir():
            bundle_path = bundle_path / default_path.name
    try:
        bundle_path.parent.mkdir(parents=True, exist_ok=True)
    except OSError as exc:
        raise wrap_io_error(
            exc,
            operation="create support bundle directory",
            path=str(bundle_path.parent),
        ) from exc
    if bundle_path.exists() and not overwrite:
        raise SupportContractError(
            f"Support bundle destination already exists: {bundle_path}",
            code="support_destination_exists",
        )
    return bundle_path


def _inventory_runtime_artifacts(repo_root: Path, cache_dir: str) -> list[dict[str, object]]:
    artifact_roots = (
        (Path(cache_dir), ".gloggur-cache"),
        (repo_root / ".gloggur" / "index", ".gloggur/index"),
    )
    artifacts: list[dict[str, object]] = []
    for root, archive_root in artifact_roots:
        if not root.exists():
            continue
        for path in sorted(root.rglob("*")):
            if not path.is_file():
                continue
            try:
                size_bytes = path.stat().st_size
            except OSError as exc:
                raise wrap_io_error(
                    exc,
                    operation="read support artifact inventory",
                    path=str(path),
                ) from exc
            artifacts.append(
                {
                    "path": f"{archive_root}/{path.relative_to(root).as_posix()}",
                    "bytes": size_bytes,
                    "source_path": str(path),
                }
            )
    return artifacts


def _public_artifact_inventory(items: list[dict[str, object]]) -> list[dict[str, object]]:
    return [
        {
            "path": item["path"],
            "bytes": item["bytes"],
        }
        for item in items
    ]


def _read_bundle_manifest(bundle_path: Path) -> dict[str, object]:
    archive_bytes = bundle_path.stat().st_size
    archive_sha256 = _sha256_file(bundle_path)
    with tarfile.open(bundle_path, "r:gz") as archive:
        manifest_member = archive.extractfile("manifest.json")
        if manifest_member is None:
            raise SupportContractError(
                "Support bundle archive is missing manifest.json.",
                code="support_session_invalid",
            )
        manifest_bytes = manifest_member.read()
    return {
        "archive_bytes": archive_bytes,
        "archive_sha256": archive_sha256,
        "manifest_sha256": sha256_bytes(manifest_bytes),
    }


def _load_capture_metadata(session_dir: Path) -> dict[str, dict[str, object]]:
    meta_path = session_dir / _CAPTURE_META_PATH
    if not meta_path.exists():
        return {}
    try:
        payload = json.loads(meta_path.read_text(encoding="utf8"))
    except (OSError, json.JSONDecodeError):
        return {}
    if not isinstance(payload, dict):
        return {}
    normalized: dict[str, dict[str, object]] = {}
    for key, value in payload.items():
        if isinstance(key, str) and isinstance(value, dict):
            normalized[key] = dict(value)
    return normalized


def _record_capture_metadata(
    *,
    session_dir: Path,
    session_rel_path: str,
    source_path: Path | None,
    source_bytes: int,
    archived_bytes: int,
    truncated: bool,
) -> None:
    payload = _load_capture_metadata(session_dir)
    record: dict[str, object] = {
        "source_bytes": source_bytes,
        "archived_bytes": archived_bytes,
        "truncated": truncated,
    }
    if source_path is not None:
        record["source_path"] = str(source_path)
    payload[session_rel_path] = record
    _write_json(session_dir / _CAPTURE_META_PATH, payload)


def _read_session_payload(session_dir: Path) -> dict[str, object]:
    try:
        payload = json.loads((session_dir / "session.json").read_text(encoding="utf8"))
    except (OSError, json.JSONDecodeError) as exc:
        raise SupportContractError(
            f"Support session metadata is invalid: {session_dir.name}",
            code="support_session_invalid",
        ) from exc
    if not isinstance(payload, dict):
        raise SupportContractError(
            f"Support session metadata is invalid: {session_dir.name}",
            code="support_session_invalid",
        )
    return payload


def _write_json(path: Path, payload: dict[str, object]) -> None:
    try:
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(
            json.dumps(payload, indent=2, sort_keys=True) + "\n",
            encoding="utf8",
        )
    except (OSError, TypeError, ValueError) as exc:
        raise wrap_io_error(
            exc, operation="write support diagnostics file", path=str(path)
        ) from exc


def _append_note(path: Path, note: str, repo_root: Path) -> None:
    sanitized = _sanitize_text(note, repo_root)
    try:
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("a", encoding="utf8") as handle:
            if path.exists() and path.stat().st_size > 0:
                handle.write("\n")
            handle.write(sanitized)
            if not sanitized.endswith("\n"):
                handle.write("\n")
    except OSError as exc:
        raise wrap_io_error(exc, operation="write support note", path=str(path)) from exc


def _sanitize_optional_string(value: str | None, repo_root: Path) -> str | None:
    if value is None:
        return None
    return _sanitize_text(value, repo_root)


def _sanitize_path(value: str, repo_root: Path) -> str:
    return _sanitize_text(value, repo_root)


def _sanitize_text(value: str, repo_root: Path) -> str:
    result = value
    repo_text = str(repo_root.resolve())
    home_text = str(Path.home().resolve())
    if repo_text:
        result = result.replace(repo_text, _REPO_ROOT_PLACEHOLDER)
    if home_text:
        result = result.replace(home_text, _HOME_PLACEHOLDER)
    return result


def _sanitize_object(value: object, repo_root: Path, *, key_name: str | None = None) -> object:
    if isinstance(value, dict):
        return {
            str(key): _sanitize_object(item, repo_root, key_name=str(key))
            for key, item in value.items()
        }
    if isinstance(value, list):
        return [_sanitize_object(item, repo_root, key_name=key_name) for item in value]
    if isinstance(value, tuple):
        return [_sanitize_object(item, repo_root, key_name=key_name) for item in value]
    if isinstance(value, str):
        if key_name and _SECRET_KEY_RE.search(key_name):
            return _REDACTED_PLACEHOLDER
        return _sanitize_text(value, repo_root)
    return value


def _parse_first_json_object(raw: str) -> dict[str, object] | None:
    start = raw.find("{")
    if start < 0:
        return None
    try:
        payload, _ = json.JSONDecoder().raw_decode(raw[start:])
    except json.JSONDecodeError:
        return None
    return payload if isinstance(payload, dict) else None


def _extract_failure_contract(payload: dict[str, object] | None) -> tuple[str | None, list[str]]:
    if not isinstance(payload, dict):
        return None, []
    child_error_code: str | None = None
    failure_codes: list[str] = []
    compatibility = payload.get("compatibility")
    if isinstance(compatibility, dict):
        nested_error = compatibility.get("error")
        if isinstance(nested_error, dict) and isinstance(nested_error.get("code"), str):
            child_error_code = str(nested_error["code"])
        nested_failure_codes = compatibility.get("failure_codes")
        if isinstance(nested_failure_codes, list):
            failure_codes = [str(code) for code in nested_failure_codes if str(code)]
    if child_error_code is None:
        error = payload.get("error")
        if isinstance(error, dict) and isinstance(error.get("code"), str):
            child_error_code = str(error["code"])
        elif isinstance(payload.get("error_code"), str):
            child_error_code = str(payload["error_code"])
    if not failure_codes:
        raw_failure_codes = payload.get("failure_codes")
        if isinstance(raw_failure_codes, list):
            failure_codes = [str(code) for code in raw_failure_codes if str(code)]
    return child_error_code, failure_codes


def _diagnostic_error_payload(exc: Exception, repo_root: Path) -> dict[str, object]:
    return {
        "captured": False,
        "error": {
            "type": type(exc).__name__,
            "detail": _sanitize_text(str(exc), repo_root),
        },
    }


def _bundle_error_payload(exc: Exception) -> dict[str, object]:
    if isinstance(exc, SupportContractError):
        return {"code": exc.code, "detail": str(exc)}
    if isinstance(exc, StorageIOError):
        return {
            "type": "io_failure",
            "category": exc.category,
            "operation": exc.operation,
            "path": exc.path,
            "detail": exc.detail,
        }
    return {"type": type(exc).__name__, "detail": str(exc)}


def _sha256_file(path: str | Path) -> str:
    try:
        return sha256_file(path)
    except OSError as exc:
        raise wrap_io_error(exc, operation="read support bundle file", path=str(path)) from exc


def _utc_now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()



from __future__ import annotations

import os
import time
from dataclasses import dataclass

from gloggur.embeddings.base import EmbeddingProvider
from gloggur.embeddings.errors import wrap_embedding_error
from gloggur.models import EdgeRecord
from gloggur.storage.metadata_store import MetadataStore


@dataclass(frozen=True)
class _RankedEdge:
    """Ranked edge wrapper used for stable sorting and payload rendering."""

    edge: EdgeRecord
    endpoint_match_score: int
    module_distance: int
    commit_penalty: int
    similarity_score: float | None = None


class GraphService:
    """Graph retrieval and semantic edge-search APIs."""

    def __init__(
        self,
        metadata_store: MetadataStore,
        embedding_provider: EmbeddingProvider | None = None,
    ) -> None:
        self.metadata_store = metadata_store
        self.embedding_provider = embedding_provider

    def neighbors(
        self,
        symbol_id: str,
        *,
        edge_type: str | None = None,
        direction: str = "both",
        k: int = 20,
    ) -> dict[str, object]:
        """Return ranked neighboring edges for a symbol endpoint."""
        start = time.time()
        edges = self.metadata_store.list_edges_for_symbol(
            symbol_id,
            direction=direction,
            edge_type=edge_type,
            limit=None,
        )
        ranked = self._rank_endpoint_edges(symbol_id=symbol_id, edges=edges, limit=k)
        duration_ms = int((time.time() - start) * 1000)
        return {
            "symbol_id": symbol_id,
            "direction": direction,
            "edge_type": edge_type,
            "results": [self._serialize_ranked_edge(item) for item in ranked],
            "metadata": {
                "total_results": len(ranked),
                "query_time_ms": duration_ms,
            },
        }

    def incoming(
        self,
        symbol_id: str,
        *,
        edge_type: str | None = None,
        k: int = 20,
    ) -> dict[str, object]:
        """Return incoming edges for a symbol."""
        return self.neighbors(symbol_id, edge_type=edge_type, direction="incoming", k=k)

    def outgoing(
        self,
        symbol_id: str,
        *,
        edge_type: str | None = None,
        k: int = 20,
    ) -> dict[str, object]:
        """Return outgoing edges for a symbol."""
        return self.neighbors(symbol_id, edge_type=edge_type, direction="outgoing", k=k)

    def search(
        self,
        query: str,
        *,
        edge_type: str | None = None,
        k: int = 20,
    ) -> dict[str, object]:
        """Semantic search over edge fact text embeddings."""
        start = time.time()
        if self.embedding_provider is None:
            raise RuntimeError("semantic graph search requires an embedding provider")
        try:
            query_vector = self.embedding_provider.embed_text(query)
        except Exception as exc:
            raise wrap_embedding_error(
                exc,
                provider=getattr(self.embedding_provider, "provider", "unknown"),
                operation="embed query for graph search",
            ) from exc
        edges = self.metadata_store.list_edges()
        if edge_type:
            edges = [edge for edge in edges if edge.edge_type == edge_type]

        ranked: list[_RankedEdge] = []
        for edge in edges:
            vector = edge.embedding_vector
            if not vector:
                continue
            if len(vector) != len(query_vector):
                continue
            distance = self._l2_distance(query_vector, vector)
            similarity_score = self._score_from_distance(distance)
            ranked.append(
                _RankedEdge(
                    edge=edge,
                    endpoint_match_score=0,
                    module_distance=0,
                    commit_penalty=0,
                    similarity_score=similarity_score,
                )
            )

        ranked.sort(
            key=lambda item: (
                -(item.similarity_score or 0.0),
                -item.edge.confidence,
                item.edge.file_path,
                item.edge.line,
                item.edge.edge_id,
            )
        )
        ranked = ranked[: max(0, k)]
        duration_ms = int((time.time() - start) * 1000)
        return {
            "query": query,
            "edge_type": edge_type,
            "results": [self._serialize_ranked_edge(item) for item in ranked],
            "metadata": {
                "total_results": len(ranked),
                "query_time_ms": duration_ms,
            },
        }

    def _rank_endpoint_edges(
        self,
        *,
        symbol_id: str,
        edges: list[EdgeRecord],
        limit: int,
    ) -> list[_RankedEdge]:
        """Apply deterministic graph ranking policy for endpoint traversal."""
        if limit < 1:
            return []
        anchor_symbol = self.metadata_store.get_symbol(symbol_id)
        anchor_path = anchor_symbol.file_path if anchor_symbol else None
        latest_commit = self._latest_commit(edges)
        ranked: list[_RankedEdge] = []
        for edge in edges:
            endpoint_match = int(edge.from_id == symbol_id) + int(edge.to_id == symbol_id)
            module_distance = self._module_distance(anchor_path, edge.file_path)
            commit_penalty = 0 if latest_commit and edge.commit == latest_commit else 1
            ranked.append(
                _RankedEdge(
                    edge=edge,
                    endpoint_match_score=endpoint_match,
                    module_distance=module_distance,
                    commit_penalty=commit_penalty,
                )
            )
        ranked.sort(
            key=lambda item: (
                -item.endpoint_match_score,
                -item.edge.confidence,
                item.module_distance,
                item.commit_penalty,
                item.edge.file_path,
                item.edge.line,
                item.edge.edge_id,
            )
        )
        return ranked[:limit]

    @staticmethod
    def _latest_commit(edges: list[EdgeRecord]) -> str | None:
        """Return a stable 'latest' commit marker for deterministic ranking."""
        commits = sorted(
            {
                edge.commit
                for edge in edges
                if isinstance(edge.commit, str) and edge.commit and edge.commit != "unknown"
            }
        )
        if not commits:
            return None
        return commits[-1]

    @staticmethod
    def _module_distance(anchor_path: str | None, other_path: str) -> int:
        """Compute directory-distance heuristic used as a ranking tie-breaker."""
        if not anchor_path:
            return 0
        anchor_parts = [
            segment
            for segment in os.path.normpath(os.path.dirname(anchor_path)).split(os.sep)
            if segment
        ]
        other_parts = [
            segment
            for segment in os.path.normpath(os.path.dirname(other_path)).split(os.sep)
            if segment
        ]
        common = 0
        for left, right in zip(anchor_parts, other_parts, strict=False):
            if left != right:
                break
            common += 1
        return (len(anchor_parts) - common) + (len(other_parts) - common)

    @staticmethod
    def _serialize_ranked_edge(item: _RankedEdge) -> dict[str, object]:
        """Render one ranked edge payload."""
        payload: dict[str, object] = {
            "edge_id": item.edge.edge_id,
            "edge_type": item.edge.edge_type,
            "from_id": item.edge.from_id,
            "to_id": item.edge.to_id,
            "from_kind": item.edge.from_kind,
            "to_kind": item.edge.to_kind,
            "file_path": item.edge.file_path,
            "line": item.edge.line,
            "confidence": item.edge.confidence,
            "repo_id": item.edge.repo_id,
            "commit": item.edge.commit,
            "text": item.edge.text,
            "ranking": {
                "endpoint_match_score": item.endpoint_match_score,
                "module_distance": item.module_distance,
                "commit_penalty": item.commit_penalty,
            },
        }
        if item.similarity_score is not None:
            payload["similarity_score"] = item.similarity_score
        return payload

    @staticmethod
    def _l2_distance(a: list[float], b: list[float]) -> float:
        """Return squared L2 distance between vectors."""
        return sum((left - right) ** 2 for left, right in zip(a, b, strict=True))

    @staticmethod
    def _score_from_distance(distance: float) -> float:
        """Convert a distance to a bounded similarity score."""
        score = 1.0 - (distance / 2.0)
        if score < 0.0:
            return 0.0
        if score > 1.0:
            return 1.0
        return score



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass(frozen=True)
class SearchIntent:
    """Public caller-facing search intent for ContextPack routing."""

    search_mode: str = "semantic"
    semantic_query: str | None = None
    ranking_mode: str = "balanced"
    result_profile: str = "default"
    semantic_assist_mode: str = "none"
    language: str | None = None
    path_prefix: str | None = None
    path_filters: tuple[str, ...] = ()
    max_files: int | None = None
    max_snippets: int | None = None
    time_budget_ms: int | None = None


@dataclass(frozen=True)
class SearchConstraints:
    """Compatibility adapter for older callers during SearchIntent migration."""

    search_mode: str = "semantic"
    language: str | None = None
    path_prefix: str | None = None
    path_filters: tuple[str, ...] = ()
    include_globs: tuple[str, ...] = ()
    exclude_globs: tuple[str, ...] = ()
    case_mode: str | None = None
    word_match: bool = False
    fixed_string: bool = False
    max_files: int | None = None
    max_snippets: int | None = None
    time_budget_ms: int | None = None


@dataclass(frozen=True)
class ExecutionHints:
    """Internal backend execution hints derived from parsed compatibility inputs."""

    include_globs: tuple[str, ...] = ()
    exclude_globs: tuple[str, ...] = ()
    case_mode: str | None = None
    word_match: bool = False
    fixed_string: bool = False


@dataclass(frozen=True)
class ContextSpan:
    """Inclusive line span for one context hit."""

    start_line: int
    end_line: int


@dataclass(frozen=True)
class ContextHit:
    """ContextPack hit consumed by agents."""

    path: str
    span: ContextSpan
    snippet: str
    score: float
    start_byte: int | None = None
    end_byte: int | None = None
    tags: tuple[str, ...] = ()

    def to_dict(self) -> dict[str, object]:
        """Serialize hit to JSON-friendly payload."""
        payload: dict[str, object] = {
            "path": self.path,
            "span": {
                "start_line": self.span.start_line,
                "end_line": self.span.end_line,
            },
            "snippet": self.snippet,
            "score": self.score,
            "tags": list(self.tags),
        }
        if self.start_byte is not None:
            payload["start_byte"] = self.start_byte
        if self.end_byte is not None:
            payload["end_byte"] = self.end_byte
        return payload


@dataclass(frozen=True)
class ContextPack:
    """Canonical search payload returned by gloggur search."""

    query: str
    summary: dict[str, object]
    hits: tuple[ContextHit, ...]
    debug: dict[str, object] | None = None
    schema_version: int = 2

    def to_dict(self, *, include_debug: bool = False) -> dict[str, object]:
        """Serialize context pack payload."""
        payload: dict[str, object] = {
            "schema_version": self.schema_version,
            "query": self.query,
            "summary": self.summary,
            "hits": [hit.to_dict() for hit in self.hits],
        }
        if include_debug and self.debug is not None:
            payload["debug"] = self.debug
        return payload


@dataclass(frozen=True)
class QueryHints:
    """Deterministic query signal extraction output."""

    symbols: tuple[str, ...] = ()
    literals: tuple[str, ...] = ()
    path_hints: tuple[str, ...] = ()
    stack_locations: tuple[tuple[str, int], ...] = ()
    identifier_tokens: tuple[str, ...] = ()
    declaration_terms: tuple[str, ...] = ()
    query_kind: str = "natural_language"


@dataclass(frozen=True)
class BackendHit:
    """Internal backend hit before pack merge/dedupe."""

    backend: str
    path: str
    start_line: int
    end_line: int
    snippet: str
    raw_score: float
    start_byte: int | None = None
    end_byte: int | None = None
    tags: tuple[str, ...] = ()


@dataclass(frozen=True)
class BackendResult:
    """Backend search execution output."""

    name: str
    hits: tuple[BackendHit, ...]
    timing_ms: int
    quality_score: float = 0.0
    error: str | None = None
    commands: tuple[str, ...] = ()


@dataclass
class RouterOutcome:
    """Final routing decision with debug-friendly metadata."""

    strategy: str
    reason: str
    winner: str | None
    considered: tuple[str, ...] = ()
    backend_scores: dict[str, float] = field(default_factory=dict)
    warning_codes: tuple[str, ...] = ()



from __future__ import annotations

import faulthandler
import json
import os
import re
import shutil
import signal
import sys
import threading
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, TextIO

from gloggur.byte_spans import discover_repo_root
from gloggur.io_failures import wrap_io_error

SUPPORT_RUNTIME_CONFIG_PATH = ".gloggur/config.toml"
DEFAULT_LOG_BYTES = 5 * 1024 * 1024
DEFAULT_RECENT_COMMAND_LIMIT = 20
HEARTBEAT_INTERVAL_SECONDS = 2.0
_REPO_ROOT_PLACEHOLDER = "<REPO_ROOT>"
_HOME_PLACEHOLDER = "<HOME>"
_REDACTED_PLACEHOLDER = "<REDACTED>"
_SECRET_KEY_RE = re.compile(r"(?:^|_)(?:key|token|secret|password)$", re.IGNORECASE)

_ACTIVE_TRACE_SESSION: CommandTraceSession | None = None


@dataclass(frozen=True)
class SupportRuntimeConfig:
    enabled: bool = False
    recent_limit: int = DEFAULT_RECENT_COMMAND_LIMIT
    max_log_bytes: int = DEFAULT_LOG_BYTES


@dataclass(frozen=True)
class SupportRuntimePaths:
    repo_root: Path
    support_root: Path
    runtime_root: Path
    active_root: Path
    recent_root: Path


class _TeeTextStream:
    def __init__(self, base: TextIO, session: CommandTraceSession, stream_name: str) -> None:
        self._base = base
        self._session = session
        self._stream_name = stream_name

    def write(self, value: str) -> int:
        written = self._base.write(value)
        self._session.append_output(self._stream_name, value)
        return written

    def flush(self) -> None:
        self._base.flush()

    def __getattr__(self, name: str) -> Any:
        return getattr(self._base, name)


class CommandTraceSession:
    def __init__(
        self,
        *,
        repo_root: Path,
        command_name: str,
        argv: list[str],
        config: SupportRuntimeConfig,
    ) -> None:
        self.repo_root = repo_root
        self.command_name = command_name
        self.argv = list(argv)
        self.config = config
        self.paths = support_runtime_paths(repo_root)
        started_at = _utc_now_iso()
        self.command_id = f"{started_at.replace(':', '').replace('-', '')}-{os.getpid()}"
        self.command_dir = self.paths.active_root / self.command_id
        self.stdout_path = self.command_dir / "stdout.log"
        self.stderr_path = self.command_dir / "stderr.log"
        self.stack_dump_path = self.command_dir / "stackdump.log"
        self.meta_path = self.command_dir / "meta.json"
        self._lock = threading.Lock()
        self._stop_event = threading.Event()
        self._heartbeat_thread: threading.Thread | None = None
        self._stdout_base: TextIO | None = None
        self._stderr_base: TextIO | None = None
        self._stack_handle: TextIO | None = None
        self._stack_signal = getattr(signal, "SIGUSR1", None)
        self._stack_supported = False
        self._installed_streams = False
        self._metadata: dict[str, object] = {
            "command_id": self.command_id,
            "command_name": self.command_name,
            "argv": _sanitize_object(self.argv, self.repo_root),
            "cwd": _sanitize_text(str(Path.cwd().resolve()), self.repo_root),
            "repo_root": _sanitize_text(str(self.repo_root.resolve()), self.repo_root),
            "pid": os.getpid(),
            "status": "running",
            "started_at": started_at,
            "updated_at": started_at,
            "heartbeat_at": started_at,
            "completed_at": None,
            "exit_code": None,
            "error_code": None,
            "failure_codes": [],
            "stage": None,
            "build_id": None,
            "build_state": None,
            "stdout_log": "stdout.log",
            "stderr_log": "stderr.log",
            "stack_dump": {
                "path": "stackdump.log",
                "supported": False,
                "signal": None,
                "last_requested_at": None,
                "last_captured_at": None,
            },
            "capture_warnings": [],
        }

    def __enter__(self) -> CommandTraceSession:
        global _ACTIVE_TRACE_SESSION
        self.command_dir.mkdir(parents=True, exist_ok=True)
        self._install_streams()
        self._install_stack_dump_handler()
        self._write_metadata()
        self._heartbeat_thread = threading.Thread(target=self._heartbeat_loop, daemon=True)
        self._heartbeat_thread.start()
        _ACTIVE_TRACE_SESSION = self
        return self

    def __exit__(self, exc_type: object, exc: object, tb: object) -> None:
        global _ACTIVE_TRACE_SESSION
        self._stop_event.set()
        if self._heartbeat_thread is not None:
            self._heartbeat_thread.join(timeout=0.2)
        self._uninstall_streams()
        self._uninstall_stack_dump_handler()
        with self._lock:
            if self._metadata.get("status") == "running":
                exit_code = int(self._metadata.get("exit_code") or 0)
                self._metadata["status"] = "completed" if exit_code == 0 else "failed"
            self._metadata["completed_at"] = _utc_now_iso()
            self._metadata["updated_at"] = _utc_now_iso()
        self._enrich_failure_from_stdout()
        self._write_metadata()
        self._move_to_recent()
        _ACTIVE_TRACE_SESSION = None

    def set_exit_code(self, exit_code: int) -> None:
        with self._lock:
            self._metadata["exit_code"] = int(exit_code)
            self._metadata["status"] = "completed" if int(exit_code) == 0 else "failed"
            self._metadata["updated_at"] = _utc_now_iso()
        self._write_metadata()

    def record_failure(
        self,
        *,
        error_code: str | None,
        failure_codes: list[str] | None = None,
        exit_code: int | None = None,
    ) -> None:
        with self._lock:
            if error_code:
                self._metadata["error_code"] = error_code
            normalized = [str(code) for code in (failure_codes or []) if str(code)]
            if normalized:
                self._metadata["failure_codes"] = normalized
            if exit_code is not None:
                self._metadata["exit_code"] = int(exit_code)
            self._metadata["status"] = "failed"
            self._metadata["updated_at"] = _utc_now_iso()
        self._write_metadata()

    def update_stage(self, stage: str, *, build_id: str | None = None) -> None:
        with self._lock:
            self._metadata["stage"] = str(stage)
            if build_id:
                self._metadata["build_id"] = str(build_id)
            self._metadata["updated_at"] = _utc_now_iso()
            self._metadata["heartbeat_at"] = _utc_now_iso()
        self._write_metadata()

    def update_build_state(self, payload: dict[str, object]) -> None:
        with self._lock:
            self._metadata["build_state"] = _sanitize_object(payload, self.repo_root)
            self._metadata["updated_at"] = _utc_now_iso()
            self._metadata["heartbeat_at"] = _utc_now_iso()
        self._write_metadata()

    def append_output(self, stream_name: str, value: str) -> None:
        if not value:
            return
        sanitized = _sanitize_text(value, self.repo_root)
        path = self.stdout_path if stream_name == "stdout" else self.stderr_path
        _append_bounded_text(path, sanitized, max_bytes=self.config.max_log_bytes)
        timestamp = _utc_now_iso()
        with self._lock:
            self._metadata["updated_at"] = timestamp
            self._metadata["heartbeat_at"] = timestamp
        self._write_metadata()

    def note_capture_warning(self, warning: str) -> None:
        with self._lock:
            warnings = list(self._metadata.get("capture_warnings", []))
            if warning not in warnings:
                warnings.append(warning)
                self._metadata["capture_warnings"] = warnings
                self._metadata["updated_at"] = _utc_now_iso()
        self._write_metadata()

    def mark_stack_dump_requested(self) -> None:
        with self._lock:
            stack_dump = dict(self._metadata.get("stack_dump", {}))
            stack_dump["last_requested_at"] = _utc_now_iso()
            self._metadata["stack_dump"] = stack_dump
            self._metadata["updated_at"] = _utc_now_iso()
        self._write_metadata()

    def mark_stack_dump_captured(self) -> None:
        with self._lock:
            stack_dump = dict(self._metadata.get("stack_dump", {}))
            stack_dump["last_captured_at"] = _utc_now_iso()
            self._metadata["stack_dump"] = stack_dump
            self._metadata["updated_at"] = _utc_now_iso()
        self._write_metadata()

    def _heartbeat_loop(self) -> None:
        while not self._stop_event.wait(HEARTBEAT_INTERVAL_SECONDS):
            with self._lock:
                if self._metadata.get("status") != "running":
                    return
                self._metadata["heartbeat_at"] = _utc_now_iso()
                self._metadata["updated_at"] = _utc_now_iso()
            self._write_metadata()

    def _install_streams(self) -> None:
        if self._installed_streams:
            return
        self._stdout_base = sys.stdout
        self._stderr_base = sys.stderr
        sys.stdout = _TeeTextStream(sys.stdout, self, "stdout")  # type: ignore[assignment]
        sys.stderr = _TeeTextStream(sys.stderr, self, "stderr")  # type: ignore[assignment]
        self._installed_streams = True

    def _uninstall_streams(self) -> None:
        if not self._installed_streams:
            return
        assert self._stdout_base is not None
        assert self._stderr_base is not None
        sys.stdout = self._stdout_base
        sys.stderr = self._stderr_base
        self._installed_streams = False

    def _install_stack_dump_handler(self) -> None:
        if os.name != "posix" or self._stack_signal is None:
            return
        try:
            self._stack_handle = self.stack_dump_path.open("a", encoding="utf8")
            faulthandler.register(self._stack_signal, file=self._stack_handle, all_threads=True)
        except (OSError, RuntimeError, ValueError):
            self._stack_supported = False
            if self._stack_handle is not None:
                self._stack_handle.close()
                self._stack_handle = None
            return
        self._stack_supported = True
        stack_dump = dict(self._metadata.get("stack_dump", {}))
        stack_dump["supported"] = True
        stack_dump["signal"] = signal.Signals(self._stack_signal).name
        self._metadata["stack_dump"] = stack_dump

    def _uninstall_stack_dump_handler(self) -> None:
        if self._stack_supported and self._stack_signal is not None:
            try:
                faulthandler.unregister(self._stack_signal)
            except RuntimeError:
                pass
        if self._stack_handle is not None:
            self._stack_handle.close()
            self._stack_handle = None

    def _enrich_failure_from_stdout(self) -> None:
        try:
            raw = self.stdout_path.read_text(encoding="utf8")
        except OSError:
            return
        payload = _parse_first_json_object(raw)
        if payload is None:
            return
        error_code, failure_codes = _extract_failure_contract(payload)
        with self._lock:
            if error_code and not self._metadata.get("error_code"):
                self._metadata["error_code"] = error_code
            existing_failure_codes = list(self._metadata.get("failure_codes", []))
            if failure_codes and not existing_failure_codes:
                self._metadata["failure_codes"] = failure_codes

    def _move_to_recent(self) -> None:
        destination = self.paths.recent_root / self.command_id
        destination.parent.mkdir(parents=True, exist_ok=True)
        if destination.exists():
            shutil.rmtree(destination)
        shutil.move(str(self.command_dir), str(destination))
        _prune_recent_commands(self.paths.recent_root, limit=self.config.recent_limit)

    def _write_metadata(self) -> None:
        _atomic_write_json(self.meta_path, self._metadata)


def support_runtime_paths(repo_root: Path) -> SupportRuntimePaths:
    support_root = repo_root / ".gloggur" / "support"
    runtime_root = support_root / "runtime"
    return SupportRuntimePaths(
        repo_root=repo_root,
        support_root=support_root,
        runtime_root=runtime_root,
        active_root=runtime_root / "active",
        recent_root=runtime_root / "recent",
    )


def load_support_runtime_config(repo_root: Path) -> SupportRuntimeConfig:
    config_path = repo_root / SUPPORT_RUNTIME_CONFIG_PATH
    if not config_path.exists():
        return SupportRuntimeConfig()
    try:
        parsed = _parse_toml(config_path)
    except OSError as exc:
        raise wrap_io_error(
            exc,
            operation="read support runtime config",
            path=str(config_path),
        ) from exc
    section = parsed.get("support")
    if not isinstance(section, dict):
        return SupportRuntimeConfig()
    enabled = bool(section.get("enabled", False))
    recent_limit = _coerce_positive_int(section.get("recent_limit"), DEFAULT_RECENT_COMMAND_LIMIT)
    max_log_bytes = _coerce_positive_int(section.get("max_log_bytes"), DEFAULT_LOG_BYTES)
    return SupportRuntimeConfig(
        enabled=enabled,
        recent_limit=recent_limit,
        max_log_bytes=max_log_bytes,
    )


def write_support_runtime_config(repo_root: Path, *, enabled: bool) -> Path:
    config_path = repo_root / SUPPORT_RUNTIME_CONFIG_PATH
    section_lines = [
        f"enabled = {'true' if enabled else 'false'}",
        f"recent_limit = {DEFAULT_RECENT_COMMAND_LIMIT}",
        f"max_log_bytes = {DEFAULT_LOG_BYTES}",
    ]
    try:
        config_path.parent.mkdir(parents=True, exist_ok=True)
        existing = config_path.read_text(encoding="utf8") if config_path.exists() else ""
        updated = _replace_toml_section(existing, "support", section_lines)
        config_path.write_text(updated, encoding="utf8")
    except OSError as exc:
        raise wrap_io_error(
            exc,
            operation="write support runtime config",
            path=str(config_path),
        ) from exc
    return config_path


def maybe_start_command_trace(
    command_name: str,
    *,
    argv: list[str] | None = None,
) -> CommandTraceSession | None:
    repo_root = discover_repo_root()
    config = load_support_runtime_config(repo_root)
    if not config.enabled:
        return None
    if command_name.startswith("support "):
        return None
    session = CommandTraceSession(
        repo_root=repo_root,
        command_name=command_name,
        argv=list(argv or sys.argv[1:]),
        config=config,
    )
    return session


def current_trace_session() -> CommandTraceSession | None:
    return _ACTIVE_TRACE_SESSION


def load_runtime_records(
    repo_root: Path,
) -> tuple[list[dict[str, object]], list[dict[str, object]]]:
    paths = support_runtime_paths(repo_root)
    active = _load_record_root(paths.active_root)
    recent = _load_record_root(paths.recent_root)
    return active, recent


def request_active_stack_dumps(repo_root: Path) -> list[str]:
    warnings: list[str] = []
    active, _recent = load_runtime_records(repo_root)
    for record in active:
        pid = record.get("pid")
        stack_dump = record.get("stack_dump")
        if not isinstance(pid, int) or pid <= 0:
            continue
        if pid == os.getpid():
            continue
        if not isinstance(stack_dump, dict) or not bool(stack_dump.get("supported")):
            continue
        signal_name = str(stack_dump.get("signal") or "")
        signal_value = getattr(signal, signal_name, None)
        if signal_value is None:
            warnings.append(f"stack_dump_signal_unsupported:{record.get('command_id')}")
            continue
        meta_path = _record_meta_path(repo_root, str(record.get("command_id")))
        stack_path = _record_stack_path(repo_root, str(record.get("command_id")))
        before_mtime = stack_path.stat().st_mtime if stack_path.exists() else 0.0
        try:
            os.kill(pid, signal_value)
        except OSError:
            warnings.append(f"stack_dump_signal_failed:{record.get('command_id')}")
            continue
        _touch_stack_request(meta_path)
        deadline = time.time() + 1.0
        while time.time() < deadline:
            if stack_path.exists() and stack_path.stat().st_mtime > before_mtime:
                _touch_stack_capture(meta_path)
                break
            time.sleep(0.05)
        else:
            warnings.append(f"stack_dump_timeout:{record.get('command_id')}")
    return warnings


def should_auto_include_cache(
    *,
    active_records: list[dict[str, object]],
    recent_records: list[dict[str, object]],
    status_payload: dict[str, object] | None,
) -> bool:
    if _status_indicates_cache_trouble(status_payload):
        return True
    for record in active_records:
        if _record_indicates_cache_trouble(record, active=True):
            return True
    for record in recent_records:
        if _record_indicates_cache_trouble(record, active=False):
            return True
    return False


def capture_enabled(repo_root: Path) -> bool:
    return load_support_runtime_config(repo_root).enabled


def _status_indicates_cache_trouble(status_payload: dict[str, object] | None) -> bool:
    if not isinstance(status_payload, dict):
        return False
    build_state = status_payload.get("build_state")
    if isinstance(build_state, dict):
        if str(build_state.get("state")) in {"building", "interrupted"}:
            return True
    warning_codes = status_payload.get("warning_codes")
    warnings = (
        set(str(code) for code in warning_codes) if isinstance(warning_codes, list) else set()
    )
    return bool(
        warnings.intersection(
            {
                "build_in_progress",
                "index_interrupted",
                "stale_build_state",
                "vector_integrity_failed",
                "chunk_span_integrity_failed",
            }
        )
    )


def _record_indicates_cache_trouble(record: dict[str, object], *, active: bool) -> bool:
    if str(record.get("command_name")) == "index":
        if active:
            return True
        exit_code = int(record.get("exit_code") or 0)
        if exit_code != 0:
            return True
    failure_codes = record.get("failure_codes")
    if isinstance(failure_codes, list):
        codes = set(str(code) for code in failure_codes)
        if codes.intersection(
            {
                "cache_lock_held",
                "index_failure",
                "stale_build_state",
                "vector_integrity_failed",
                "chunk_span_integrity_failed",
            }
        ):
            return True
    build_state = record.get("build_state")
    if isinstance(build_state, dict) and str(build_state.get("state")) in {
        "building",
        "interrupted",
    }:
        return True
    return False


def _record_meta_path(repo_root: Path, command_id: str) -> Path:
    paths = support_runtime_paths(repo_root)
    for root in (paths.active_root, paths.recent_root):
        candidate = root / command_id / "meta.json"
        if candidate.exists():
            return candidate
    return paths.active_root / command_id / "meta.json"


def _record_stack_path(repo_root: Path, command_id: str) -> Path:
    paths = support_runtime_paths(repo_root)
    for root in (paths.active_root, paths.recent_root):
        candidate = root / command_id / "stackdump.log"
        if candidate.exists():
            return candidate
    return paths.active_root / command_id / "stackdump.log"


def _touch_stack_request(meta_path: Path) -> None:
    if not meta_path.exists():
        return
    try:
        payload = json.loads(meta_path.read_text(encoding="utf8"))
    except (OSError, json.JSONDecodeError):
        return
    if not isinstance(payload, dict):
        return
    stack_dump = dict(payload.get("stack_dump", {}))
    stack_dump["last_requested_at"] = _utc_now_iso()
    payload["stack_dump"] = stack_dump
    _atomic_write_json(meta_path, payload)


def _touch_stack_capture(meta_path: Path) -> None:
    if not meta_path.exists():
        return
    try:
        payload = json.loads(meta_path.read_text(encoding="utf8"))
    except (OSError, json.JSONDecodeError):
        return
    if not isinstance(payload, dict):
        return
    stack_dump = dict(payload.get("stack_dump", {}))
    stack_dump["last_captured_at"] = _utc_now_iso()
    payload["stack_dump"] = stack_dump
    _atomic_write_json(meta_path, payload)


def _load_record_root(root: Path) -> list[dict[str, object]]:
    if not root.exists():
        return []
    payloads: list[dict[str, object]] = []
    for command_dir in sorted(root.iterdir()):
        meta_path = command_dir / "meta.json"
        if not meta_path.is_file():
            continue
        try:
            payload = json.loads(meta_path.read_text(encoding="utf8"))
        except (OSError, json.JSONDecodeError):
            continue
        if isinstance(payload, dict):
            payloads.append(payload)
    payloads.sort(
        key=lambda item: str(item.get("updated_at") or item.get("started_at") or ""),
        reverse=True,
    )
    return payloads


def _prune_recent_commands(root: Path, *, limit: int) -> None:
    if not root.exists():
        return
    command_dirs = sorted(
        (path for path in root.iterdir() if path.is_dir()),
        key=lambda path: path.stat().st_mtime,
        reverse=True,
    )
    for path in command_dirs[limit:]:
        shutil.rmtree(path, ignore_errors=True)


def _append_bounded_text(path: Path, value: str, *, max_bytes: int) -> None:
    payload = value.encode("utf8", errors="replace")
    path.parent.mkdir(parents=True, exist_ok=True)
    with path.open("ab") as handle:
        handle.write(payload)
    try:
        if path.stat().st_size <= max_bytes:
            return
        raw = path.read_bytes()
        path.write_bytes(raw[-max_bytes:])
    except OSError as exc:
        raise wrap_io_error(exc, operation="write support runtime log", path=str(path)) from exc


def _atomic_write_json(path: Path, payload: dict[str, object]) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    temp_path = path.with_suffix(path.suffix + ".tmp")
    try:
        temp_path.write_text(json.dumps(payload, indent=2, sort_keys=True) + "\n", encoding="utf8")
        temp_path.replace(path)
    except OSError as exc:
        raise wrap_io_error(
            exc,
            operation="write support runtime metadata",
            path=str(path),
        ) from exc


def _parse_toml(path: Path) -> dict[str, object]:
    try:
        import tomllib  # type: ignore[attr-defined]

        with path.open("rb") as handle:
            parsed = tomllib.load(handle)
    except ModuleNotFoundError:
        import tomli  # type: ignore[import-not-found]

        with path.open("rb") as handle:
            parsed = tomli.load(handle)
    if isinstance(parsed, dict):
        return parsed
    return {}


def _replace_toml_section(existing: str, section_name: str, body_lines: list[str]) -> str:
    header = f"[{section_name}]"
    lines = existing.splitlines()
    output: list[str] = []
    index = 0
    replaced = False
    while index < len(lines):
        line = lines[index]
        if line.strip() == header:
            replaced = True
            output.append(header)
            output.extend(body_lines)
            index += 1
            while index < len(lines) and not lines[index].startswith("["):
                index += 1
            continue
        output.append(line)
        index += 1
    if not replaced:
        if output and output[-1] != "":
            output.append("")
        output.append(header)
        output.extend(body_lines)
    return "\n".join(output).rstrip() + "\n"


def _coerce_positive_int(value: object, default: int) -> int:
    try:
        candidate = int(value)
    except (TypeError, ValueError):
        return default
    return candidate if candidate > 0 else default


def _sanitize_text(value: str, repo_root: Path) -> str:
    result = value
    repo_text = str(repo_root.resolve())
    home_text = str(Path.home().resolve())
    if repo_text:
        result = result.replace(repo_text, _REPO_ROOT_PLACEHOLDER)
    if home_text:
        result = result.replace(home_text, _HOME_PLACEHOLDER)
    return result


def _sanitize_object(value: object, repo_root: Path, *, key_name: str | None = None) -> object:
    if isinstance(value, dict):
        return {
            str(key): _sanitize_object(item, repo_root, key_name=str(key))
            for key, item in value.items()
        }
    if isinstance(value, list):
        return [_sanitize_object(item, repo_root, key_name=key_name) for item in value]
    if isinstance(value, tuple):
        return [_sanitize_object(item, repo_root, key_name=key_name) for item in value]
    if isinstance(value, str):
        if key_name and _SECRET_KEY_RE.search(key_name):
            return _REDACTED_PLACEHOLDER
        return _sanitize_text(value, repo_root)
    return value


def _parse_first_json_object(raw: str) -> dict[str, object] | None:
    start = raw.find("{")
    if start < 0:
        return None
    try:
        payload, _ = json.JSONDecoder().raw_decode(raw[start:])
    except json.JSONDecodeError:
        return None
    return payload if isinstance(payload, dict) else None


def _extract_failure_contract(payload: dict[str, object] | None) -> tuple[str | None, list[str]]:
    if not isinstance(payload, dict):
        return None, []
    error_code: str | None = None
    failure_codes: list[str] = []
    error = payload.get("error")
    if isinstance(error, dict) and isinstance(error.get("code"), str):
        error_code = str(error["code"])
    elif isinstance(payload.get("error_code"), str):
        error_code = str(payload["error_code"])
    raw_failure_codes = payload.get("failure_codes")
    if isinstance(raw_failure_codes, list):
        failure_codes = [str(code) for code in raw_failure_codes if str(code)]
    compatibility = payload.get("compatibility")
    if isinstance(compatibility, dict) and not failure_codes:
        nested = compatibility.get("failure_codes")
        if isinstance(nested, list):
            failure_codes = [str(code) for code in nested if str(code)]
        nested_error = compatibility.get("error")
        if isinstance(nested_error, dict) and isinstance(nested_error.get("code"), str):
            error_code = str(nested_error["code"])
    return error_code, failure_codes


def _utc_now_iso() -> str:
    return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())



from __future__ import annotations

import hashlib
import os
import re
from collections.abc import Iterable
from dataclasses import dataclass

from tree_sitter import Node, Tree
from tree_sitter_language_pack import get_parser

from gloggur.models import Symbol
from gloggur.parsers.base import ParsedFile, Parser
from gloggur.parsers.signal_processors import (
    ParserSignalProcessor,
    SignalProcessingOutcome,
    default_signal_processors,
    project_legacy_fields,
)


@dataclass(frozen=True)
class LanguageSpec:
    """Tree-sitter node type configuration for functions/classes/interfaces."""

    name: str
    function_nodes: list[str]
    class_nodes: list[str]
    interface_nodes: list[str]


_LANGUAGE_SPECS: dict[str, LanguageSpec] = {
    "python": LanguageSpec(
        "python",
        ["function_definition", "decorated_definition"],
        ["class_definition"],
        [],
    ),
    "javascript": LanguageSpec(
        "javascript",
        ["function_declaration", "generator_function_declaration"],
        ["class_declaration"],
        [],
    ),
    "typescript": LanguageSpec(
        "typescript",
        ["function_declaration"],
        ["class_declaration"],
        ["interface_declaration"],
    ),
    "tsx": LanguageSpec(
        "tsx",
        ["function_declaration"],
        ["class_declaration"],
        ["interface_declaration"],
    ),
    "rust": LanguageSpec(
        "rust",
        ["function_item"],
        ["struct_item", "enum_item"],
        ["trait_item"],
    ),
    "go": LanguageSpec(
        "go",
        ["function_declaration", "method_declaration"],
        [],
        ["interface_type"],
    ),
    "java": LanguageSpec(
        "java",
        ["method_declaration", "constructor_declaration"],
        ["class_declaration"],
        ["interface_declaration"],
    ),
}


@dataclass(frozen=True)
class _ContainerContext:
    """Container stack frame for fqname/container metadata."""

    symbol_id: str
    fqname: str


class TreeSitterParser(Parser):
    """Tree-sitter parser that extracts Symbol objects from source code."""

    def __init__(
        self,
        language: str,
        signal_processors: list[ParserSignalProcessor] | None = None,
    ) -> None:
        """Initialize a tree-sitter parser for the given language."""
        if language not in _LANGUAGE_SPECS:
            raise ValueError(f"Unsupported language: {language}")
        self.language = language
        self.spec = _LANGUAGE_SPECS[language]
        self.parser = get_parser(language)
        self.signal_processors = signal_processors or default_signal_processors()

    def parse_file(self, path: str, source: str) -> ParsedFile:
        """Parse a file and return ParsedFile with extracted symbols."""
        symbols = self.extract_symbols(path, source)
        return ParsedFile(path=path, language=self.language, source=source, symbols=symbols)

    def parse_with_edit(self, old_tree: Tree | None, new_source: str) -> Tree:
        """Parse source with optional incremental edit support."""
        return self.parser.parse(bytes(new_source, "utf8"), old_tree)

    def extract_symbols(self, path: str, source: str) -> list[Symbol]:
        """Extract symbol records from the syntax tree with container metadata."""
        tree = self.parser.parse(bytes(source, "utf8"))
        repo_id = self._repo_id(path)
        symbols: list[Symbol] = []
        self._collect_symbols(
            node=tree.root_node,
            path=path,
            source=source,
            repo_id=repo_id,
            containers=[],
            out=symbols,
        )
        symbols.sort(key=lambda item: (item.file_path, item.start_line, item.end_line, item.id))
        return symbols

    def get_supported_languages(self) -> Iterable[str]:
        """Return the single language supported by this parser."""
        return [self.language]

    def _collect_symbols(
        self,
        *,
        node: Node,
        path: str,
        source: str,
        repo_id: str,
        containers: list[_ContainerContext],
        out: list[Symbol],
    ) -> None:
        """Depth-first symbol extraction with container stack propagation."""
        if not node.is_named:
            return

        current_container = containers[-1] if containers else None
        symbol = self._symbol_from_node(
            node,
            path=path,
            source=source,
            repo_id=repo_id,
            container=current_container,
        )
        pushed = False
        if symbol is not None:
            out.append(symbol)
            fqname = symbol.fqname or symbol.name
            containers.append(_ContainerContext(symbol_id=symbol.id, fqname=fqname))
            pushed = True

        for child in node.children:
            self._collect_symbols(
                node=child,
                path=path,
                source=source,
                repo_id=repo_id,
                containers=containers,
                out=out,
            )

        if pushed:
            containers.pop()

    def _symbol_from_node(
        self,
        node: Node,
        *,
        path: str,
        source: str,
        repo_id: str,
        container: _ContainerContext | None,
    ) -> Symbol | None:
        """Convert one AST node into a Symbol with fqname/container metadata."""
        kind = self._symbol_kind(node)
        if not kind:
            return None
        name = self._extract_name(node, source)
        if not name:
            return None

        signals = []
        attributes: dict[str, object] = {}
        final_kind = kind
        for processor in self.signal_processors:
            outcome = processor.process(
                language=self.language,
                node=node,
                name=name,
                kind=kind,
                source=source,
            )
            if not isinstance(outcome, SignalProcessingOutcome):
                continue
            if outcome.kind_override:
                final_kind = outcome.kind_override
            if outcome.signals:
                signals.extend(outcome.signals)
            if outcome.attributes:
                attributes.update(outcome.attributes)

        (
            invariants,
            calls,
            is_serialization_boundary,
            implicit_contract,
        ) = project_legacy_fields(signals)

        signature = self._extract_signature(node, source)
        docstring = self._extract_docstring(node, source)
        body = source[node.start_byte : node.end_byte]
        body_hash = hashlib.sha256(body.encode("utf8")).hexdigest()
        start_line = node.start_point[0] + 1
        end_line = node.end_point[0] + 1
        fqname = name if container is None else f"{container.fqname}.{name}"
        visibility, exported = self._infer_visibility_and_exported(
            node=node,
            name=name,
            signature=signature,
            source=source,
            container=container,
        )
        symbol_id = self._hash_symbol_id(
            repo_id=repo_id,
            path=path,
            language=self.language,
            kind=final_kind,
            fqname=fqname,
            start_line=start_line,
            end_line=end_line,
        )
        return Symbol(
            id=symbol_id,
            name=name,
            kind=final_kind,
            fqname=fqname,
            file_path=path,
            start_line=start_line,
            end_line=end_line,
            container_id=container.symbol_id if container else None,
            container_fqname=container.fqname if container else None,
            signature=signature,
            docstring=docstring,
            body_hash=body_hash,
            language=self.language,
            repo_id=repo_id,
            commit=None,
            visibility=visibility,
            exported=exported,
            tokens_estimate=self._estimate_tokens(body),
            invariants=invariants,
            calls=calls,
            is_serialization_boundary=is_serialization_boundary,
            implicit_contract=implicit_contract,
            signals=signals,
            attributes=attributes,
        )

    def _symbol_kind(self, node: Node) -> str | None:
        """Classify a node as function/method/class/interface/trait/enum/type."""
        if self._is_decorated_inner_symbol(node):
            return None

        if self.language == "python":
            if node.type == "function_definition":
                parent = node.parent
                if parent and parent.type == "class_definition":
                    return "method"
                return "function"
            if node.type == "decorated_definition":
                child_kind = self._decorated_child_kind(node)
                return child_kind
            if node.type == "class_definition":
                return "class"
            return None

        if self.language in {"javascript", "typescript", "tsx"}:
            if node.type == "method_definition":
                return "method"
            if node.type in self.spec.function_nodes:
                return "function"
            if node.type in self.spec.class_nodes:
                return "class"
            if node.type in self.spec.interface_nodes:
                return "interface"
            return None

        if self.language == "go":
            if node.type == "method_declaration":
                return "method"
            if node.type == "function_declaration":
                return "function"
            if node.type == "interface_type":
                return "interface"
            return None

        if self.language == "rust":
            if node.type == "function_item":
                parent = node.parent
                if parent and parent.type in {"impl_item", "trait_item"}:
                    return "method"
                return "function"
            if node.type == "struct_item":
                return "type"
            if node.type == "enum_item":
                return "enum"
            if node.type == "trait_item":
                return "trait"
            return None

        if self.language == "java":
            if node.type in {"method_declaration", "constructor_declaration"}:
                return "method"
            if node.type == "class_declaration":
                return "class"
            if node.type == "interface_declaration":
                return "interface"
            return None

        if node.type in self.spec.function_nodes:
            return "function"
        if node.type in self.spec.class_nodes:
            return "class"
        if node.type in self.spec.interface_nodes:
            return "interface"
        return None

    def _is_decorated_inner_symbol(self, node: Node) -> bool:
        """Return True when node is the inner declaration under decorated_definition."""
        if self.language != "python":
            return False
        parent = node.parent
        if not parent or parent.type != "decorated_definition":
            return False
        return node.type in {"function_definition", "class_definition"}

    def _decorated_child_kind(self, node: Node) -> str | None:
        """Resolve symbol kind for Python decorated definitions."""
        for child in node.children:
            if child.type == "function_definition":
                parent = child.parent
                if parent and parent.parent and parent.parent.type == "class_definition":
                    return "method"
                return "function"
            if child.type == "class_definition":
                return "class"
        return None

    def _extract_name(self, node: Node, source: str) -> str | None:
        """Extract the symbol name from a node."""
        if node.type == "decorated_definition":
            for child in node.children:
                if child.type in self.spec.function_nodes and child.type != "decorated_definition":
                    return self._extract_name(child, source)
                if child.type in self.spec.class_nodes:
                    return self._extract_name(child, source)

        for child in node.children:
            if child.type in {
                "identifier",
                "type_identifier",
                "name",
                "property_identifier",
                "field_identifier",
            }:
                return source[child.start_byte : child.end_byte]
        return None

    def _extract_signature(self, node: Node, source: str) -> str | None:
        """Extract the first-line signature for a symbol."""
        if node.type == "decorated_definition":
            for child in node.children:
                if child.type in self.spec.function_nodes and child.type != "decorated_definition":
                    return self._extract_signature(child, source)
                if child.type in self.spec.class_nodes:
                    return self._extract_signature(child, source)

        text = source[node.start_byte : node.end_byte]
        first_line = text.splitlines()[0] if text else ""
        return first_line.strip() or None

    def _extract_docstring(self, node: Node, source: str) -> str | None:
        """Extract a docstring or nearby comment for a symbol."""
        if node.type == "decorated_definition":
            for child in node.children:
                if child.type in self.spec.function_nodes and child.type != "decorated_definition":
                    return self._extract_docstring(child, source)
                if child.type in self.spec.class_nodes:
                    return self._extract_docstring(child, source)

        if self.language == "python":
            body = None
            for child in node.named_children:
                if child.type == "block":
                    body = child
                    break
            if body and body.named_child_count:
                first = body.named_children[0]
                if first.type in {"string", "string_literal"}:
                    return self._strip_quotes(source[first.start_byte : first.end_byte])
                if first.type == "expression_statement" and first.named_child_count:
                    expr = first.named_children[0]
                    if expr.type in {"string", "string_literal"}:
                        return self._strip_quotes(source[expr.start_byte : expr.end_byte])
        prev = node.prev_named_sibling
        if prev and "comment" in prev.type:
            return source[prev.start_byte : prev.end_byte].strip()
        return None

    def _infer_visibility_and_exported(
        self,
        *,
        node: Node,
        name: str,
        signature: str | None,
        source: str,
        container: _ContainerContext | None,
    ) -> tuple[str | None, bool | None]:
        """Infer visibility/export status using language-specific heuristics."""
        _ = container
        sig = (signature or "").strip()
        full_text = source[node.start_byte : min(node.end_byte, node.start_byte + 256)]

        if self.language == "python":
            private = name.startswith("_")
            return ("private" if private else "public", not private)

        if self.language == "go":
            exported = bool(name) and name[0].isupper()
            return ("public" if exported else "private", exported)

        if self.language == "rust":
            exported = bool(re.search(r"\bpub\b", sig))
            return ("public" if exported else "private", exported)

        if self.language in {"javascript", "typescript", "tsx", "java"}:
            lowered = f"{sig} {full_text}".lower()
            if " private " in f" {lowered} ":
                return "private", False
            if " protected " in f" {lowered} ":
                return "protected", False
            if " public " in f" {lowered} ":
                return "public", True
            if " export " in f" {lowered} ":
                return "public", True
            # Best effort fallback: class/module scoped declarations default to public.
            return "public", True

        return None, None

    @staticmethod
    def _estimate_tokens(text: str) -> int:
        """Estimate token count for sizing/ranking constraints."""
        stripped = text.strip()
        if not stripped:
            return 0
        return max(1, len(stripped.split()))

    @staticmethod
    def _repo_id(path: str) -> str:
        """Return deterministic repo id hash from nearest .git root or file parent."""
        absolute = os.path.abspath(path)
        current = absolute if os.path.isdir(absolute) else os.path.dirname(absolute)
        while True:
            git_path = os.path.join(current, ".git")
            if os.path.exists(git_path):
                return hashlib.sha256(current.encode("utf8")).hexdigest()
            parent = os.path.dirname(current)
            if parent == current:
                break
            current = parent
        fallback = os.path.dirname(absolute) if os.path.isfile(absolute) else absolute
        return hashlib.sha256(fallback.encode("utf8")).hexdigest()

    @staticmethod
    def _hash_symbol_id(
        *,
        repo_id: str,
        path: str,
        language: str,
        kind: str,
        fqname: str,
        start_line: int,
        end_line: int,
    ) -> str:
        """Build stable symbol id from the Spec 1 hash contract."""
        payload = "|".join(
            [
                repo_id,
                os.path.normpath(path),
                language,
                kind,
                fqname,
                str(start_line),
                str(end_line),
            ]
        )
        return hashlib.sha256(payload.encode("utf8")).hexdigest()

    @staticmethod
    def _strip_quotes(value: str) -> str:
        """Strip surrounding quotes from a string literal."""
        if value.startswith(("'''", '"""')) and value.endswith(("'''", '"""')):
            return value[3:-3].strip()
        if value.startswith(("'", '"')) and value.endswith(("'", '"')):
            return value[1:-1].strip()
        return value.strip()



from __future__ import annotations

import io
import sys
from collections.abc import Iterable
from contextlib import redirect_stderr, redirect_stdout

from gloggur.embeddings.base import EmbeddingProvider

_EXPECTED_WRAPPER_WARNING_TOKENS = (
    "No sentence-transformers model found",
    "Creating a new one with mean pooling.",
)
_EXPECTED_BOOTSTRAP_PROGRESS_TOKENS = (
    "Loading weights:",
    "Materializing param=",
)


class LocalEmbeddingProvider(EmbeddingProvider):
    """Embedding provider backed by a local sentence-transformers model."""

    def __init__(
        self,
        model_name: str,
        cache_dir: str | None = None,
        fallback_cache_dir: str | None = None,
    ) -> None:
        """Configure local model loading."""
        self.provider = "local"
        self.model_name = model_name
        self.cache_dir = cache_dir
        self.fallback_cache_dir = fallback_cache_dir
        self._model = None

    def _load_model(self):
        """Load the sentence-transformers model."""
        if self._model is not None:
            return self._model
        try:
            from sentence_transformers import SentenceTransformer
        except ImportError as exc:
            raise RuntimeError(
                "sentence-transformers is required for local embeddings; "
                "install local extras (pip install -e '.[local]')."
            ) from exc
        try:
            stdout_buffer = io.StringIO()
            stderr_buffer = io.StringIO()
            with redirect_stdout(stdout_buffer), redirect_stderr(stderr_buffer):
                self._model = SentenceTransformer(self.model_name, cache_folder=self.cache_dir)
        except Exception as exc:
            raise RuntimeError(
                f"Failed to load local embedding model '{self.model_name}' "
                f"({type(exc).__name__}: {exc})"
            ) from exc
        filtered_stdout = self._filter_expected_bootstrap_output(stdout_buffer.getvalue())
        filtered_stderr = self._filter_expected_bootstrap_output(stderr_buffer.getvalue())
        unexpected_output = f"{filtered_stdout}{filtered_stderr}"
        if unexpected_output:
            sys.stderr.write(unexpected_output)
        return self._model

    @staticmethod
    def _filter_expected_bootstrap_output(output_text: str) -> str:
        """Suppress expected bootstrap chatter while preserving actionable diagnostics."""
        if not output_text:
            return ""
        kept_lines: list[str] = []
        for line in output_text.splitlines(keepends=True):
            stripped = line.lstrip()
            if not stripped.strip():
                continue
            if all(token in stripped for token in _EXPECTED_WRAPPER_WARNING_TOKENS):
                continue
            if any(token in stripped for token in _EXPECTED_BOOTSTRAP_PROGRESS_TOKENS):
                continue
            kept_lines.append(line)
        return "".join(kept_lines)

    @staticmethod
    def _filter_expected_wrapper_warning(stderr_text: str) -> str:
        """Suppress only the expected sentence-transformers wrapper bootstrap line."""
        return LocalEmbeddingProvider._filter_expected_bootstrap_output(stderr_text)

    def embed_text(self, text: str) -> list[float]:
        """Embed a single text string."""
        model = self._load_model()
        vector = model.encode([text], normalize_embeddings=True)[0]
        return vector.tolist()

    def embed_batch(self, texts: Iterable[str]) -> list[list[float]]:
        """Embed a batch of text strings."""
        payload = list(texts)
        if not payload:
            return []
        model = self._load_model()
        vectors = model.encode(payload, normalize_embeddings=True)
        return [vector.tolist() for vector in vectors]

    def get_dimension(self) -> int:
        """Return the embedding dimension."""
        model = self._load_model()
        return model.get_sentence_embedding_dimension()



"""Structured I/O failure classification and error types for cache operations.

Provides ``StorageIOError``, a stable dataclass exception that wraps raw
``OSError`` and ``sqlite3`` exceptions with a machine-readable category, a
probable-cause description, and actionable remediation steps.  All cache write
paths should raise ``StorageIOError`` so that CLI outputs can emit consistent
JSON payloads regardless of the underlying OS or database error.
"""

from __future__ import annotations

import errno
import sqlite3
from dataclasses import dataclass

IO_ERROR_CATEGORIES = {
    "cache_lock_held",
    "permission_denied",
    "read_only_filesystem",
    "disk_full_or_quota",
    "path_not_writable",
    "unknown_io_error",
}

_CATEGORY_DETAILS: dict[str, tuple[str, list[str]]] = {
    "cache_lock_held": (
        "Another gloggur process currently holds the cache writer lock.",
        [
            "Wait for the active `index`, `watch`, or `clear-cache` command to finish.",
            "Retry after the lock holder exits or releases the cache write lock.",
        ],
    ),
    "permission_denied": (
        "The process does not have permission for this filesystem operation.",
        [
            "Check path ownership and filesystem permissions.",
            "Run with a user that can read/write the cache directory.",
        ],
    ),
    "read_only_filesystem": (
        "The target path is on a read-only filesystem.",
        [
            "Use a writable cache directory (set GLOGGUR_CACHE_DIR).",
            "If appropriate, remount the filesystem as read-write.",
        ],
    ),
    "disk_full_or_quota": (
        "The filesystem is out of space or quota for writes.",
        [
            "Free disk space or increase quota.",
            "Retry the command after capacity is restored.",
        ],
    ),
    "path_not_writable": (
        "The target path or one of its parent directories is not writable/usable.",
        [
            "Ensure the parent directory exists and is writable.",
            "Set GLOGGUR_CACHE_DIR to a valid writable path.",
        ],
    ),
    "unknown_io_error": (
        "An unclassified filesystem or database I/O failure occurred.",
        [
            "Inspect the original error detail for OS-specific context.",
            "Retry with a known-good writable cache directory.",
        ],
    ),
}

_PERMISSION_ERRNOS = {errno.EACCES, errno.EPERM}
_READ_ONLY_ERRNOS = {errno.EROFS}
_DISK_ERRNOS = {errno.ENOSPC}
_PATH_ERRNOS = {
    errno.ENOENT,
    errno.ENOTDIR,
    errno.EISDIR,
    errno.ENAMETOOLONG,
    errno.ELOOP,
}
if hasattr(errno, "EDQUOT"):
    _DISK_ERRNOS.add(errno.EDQUOT)


@dataclass
class StorageIOError(RuntimeError):
    """Structured I/O failure for deterministic user/CI handling."""

    category: str
    operation: str
    path: str
    probable_cause: str
    remediation: list[str]
    detail: str
    context: dict[str, object] | None = None

    def __post_init__(self) -> None:
        """Validate that ``category`` is one of the known stable categories."""
        if self.category not in IO_ERROR_CATEGORIES:
            raise ValueError(f"unsupported io error category: {self.category}")

    def __str__(self) -> str:
        """Return a single-line summary suitable for log output and exception chains."""
        return (
            f"IO failure [{self.category}] during '{self.operation}' at {self.path}: "
            f"{self.detail}"
        )

    def to_payload(self) -> dict[str, object]:
        """Return machine-readable error payload for JSON outputs."""
        error: dict[str, object] = {
            "type": "io_failure",
            "category": self.category,
            "operation": self.operation,
            "path": self.path,
            "probable_cause": self.probable_cause,
            "remediation": self.remediation,
            "detail": self.detail,
        }
        if isinstance(self.context, dict):
            error.update(self.context)
        return {"error": error}


def wrap_io_error(exc: Exception, *, operation: str, path: str) -> StorageIOError:
    """Convert an OSError/sqlite exception into a structured StorageIOError."""
    if isinstance(exc, StorageIOError):
        return exc
    category = classify_io_error(exc)
    probable_cause, remediation = _CATEGORY_DETAILS[category]
    detail = f"{type(exc).__name__}: {exc}"
    return StorageIOError(
        category=category,
        operation=operation,
        path=path,
        probable_cause=probable_cause,
        remediation=list(remediation),
        detail=detail,
    )


def classify_io_error(exc: Exception) -> str:
    """Classify an exception into one of the stable IO error categories."""
    if isinstance(exc, OSError):
        return _classify_os_error(exc)
    if isinstance(exc, sqlite3.OperationalError):
        return _classify_sqlite_operational_error(exc)
    if isinstance(exc, sqlite3.DatabaseError):
        return _classify_sqlite_database_error(exc)
    return "unknown_io_error"


def _classify_os_error(exc: OSError) -> str:
    """Map an ``OSError`` to a stable IO error category string."""
    if isinstance(exc, PermissionError) or exc.errno in _PERMISSION_ERRNOS:
        return "permission_denied"
    if exc.errno in _READ_ONLY_ERRNOS:
        return "read_only_filesystem"
    if exc.errno in _DISK_ERRNOS:
        return "disk_full_or_quota"
    if exc.errno in _PATH_ERRNOS:
        return "path_not_writable"
    return "unknown_io_error"


def _classify_sqlite_operational_error(exc: sqlite3.OperationalError) -> str:
    """Map a ``sqlite3.OperationalError`` to a stable IO error category."""
    detail = str(exc).lower()
    return _classify_sqlite_error_detail(detail)


def _classify_sqlite_database_error(exc: sqlite3.DatabaseError) -> str:
    """Map a generic ``sqlite3.DatabaseError`` to a stable IO error category."""
    detail = str(exc).lower()
    return _classify_sqlite_error_detail(detail)


def _classify_sqlite_error_detail(detail: str) -> str:
    """Classify a lowercased sqlite error message into a stable IO error category."""
    if any(token in detail for token in ("readonly", "read-only")):
        return "read_only_filesystem"
    if "permission denied" in detail:
        return "permission_denied"
    if any(token in detail for token in ("disk is full", "database or disk is full", "quota")):
        return "disk_full_or_quota"
    if "unable to open database file" in detail:
        return "path_not_writable"
    if "i/o error" in detail:
        return "unknown_io_error"
    return "unknown_io_error"


def format_io_error_message(error: StorageIOError) -> str:
    """Create stable human-readable stderr output for I/O failures."""
    lines = [
        str(error),
        f"Probable cause: {error.probable_cause}",
        "Remediation:",
    ]
    lines.extend(f"{idx}. {step}" for idx, step in enumerate(error.remediation, start=1))
    return "\n".join(lines)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable


class EmbeddingProvider(ABC):
    """Abstract interface for embedding providers."""

    @abstractmethod
    def embed_text(self, text: str) -> list[float]:
        """Embed a single text string."""
        raise NotImplementedError

    @abstractmethod
    def embed_batch(self, texts: Iterable[str]) -> list[list[float]]:
        """Embed a batch of text strings."""
        raise NotImplementedError

    @abstractmethod
    def get_dimension(self) -> int:
        """Return the vector dimensionality for embeddings."""
        raise NotImplementedError



from __future__ import annotations

import os
import sqlite3
from collections.abc import Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

from gloggur.symbol_index.models import IndexedFile, SymbolOccurrence

SQLITE_BUSY_TIMEOUT_MS = 5_000
SQLITE_CONNECT_TIMEOUT_SECONDS = SQLITE_BUSY_TIMEOUT_MS / 1000
SQLITE_JOURNAL_MODE = "WAL"
SQLITE_SYNCHRONOUS = "NORMAL"
SCHEMA_VERSION_KEY = "schema_version"
SYMBOL_INDEX_SCHEMA_VERSION = "2"
REQUIRED_TABLES = {"occurrences", "files", "meta"}
REQUIRED_COLUMNS = {
    "occurrences": {
        "symbol",
        "kind",
        "path",
        "start_line",
        "end_line",
        "start_byte",
        "end_byte",
        "language",
        "container",
        "signature",
    },
    "files": {"path", "content_hash", "mtime_ns", "language", "last_indexed"},
    "meta": {"key", "value"},
}


@dataclass(frozen=True)
class SymbolIndexStoreConfig:
    repo_root: Path

    @property
    def db_path(self) -> Path:
        return self.repo_root / ".gloggur" / "index" / "symbols.db"


class SymbolIndexStore:
    """SQLite-backed symbol occurrence index at .gloggur/index/symbols.db."""

    def __init__(self, config: SymbolIndexStoreConfig, *, create_if_missing: bool = True) -> None:
        self.config = config
        self._create_if_missing = create_if_missing
        self._available = True
        self._unavailability_reason: str | None = None
        self.last_reset_reason: str | None = None
        db_path = self.config.db_path
        if not create_if_missing and not db_path.exists():
            self._available = False
            self._unavailability_reason = "missing or unreadable symbols.db"
            return
        db_path.parent.mkdir(parents=True, exist_ok=True)
        schema_problem = self._schema_problem() if db_path.exists() else None
        if schema_problem:
            if create_if_missing:
                self._reset_database(schema_problem)
            else:
                self._available = False
                self._unavailability_reason = schema_problem
                return
        if create_if_missing or db_path.exists():
            self._init_db()

    @property
    def db_path(self) -> str:
        return str(self.config.db_path)

    @property
    def available(self) -> bool:
        return self._available and self.config.db_path.exists()

    @property
    def unavailability_reason(self) -> str | None:
        return self._unavailability_reason

    @contextmanager
    def _connect(self) -> Iterator[sqlite3.Connection]:
        if not self.available and not self._create_if_missing:
            raise RuntimeError("symbol index database is unavailable")
        conn = sqlite3.connect(str(self.config.db_path), timeout=SQLITE_CONNECT_TIMEOUT_SECONDS)
        conn.row_factory = sqlite3.Row
        try:
            conn.execute(f"PRAGMA busy_timeout = {SQLITE_BUSY_TIMEOUT_MS}")
            conn.execute(f"PRAGMA journal_mode = {SQLITE_JOURNAL_MODE}")
            conn.execute(f"PRAGMA synchronous = {SQLITE_SYNCHRONOUS}")
            yield conn
            conn.commit()
        except Exception:
            try:
                conn.rollback()
            except sqlite3.OperationalError:
                pass
            raise
        finally:
            conn.close()

    def _init_db(self) -> None:
        with self._connect() as conn:
            conn.executescript("""
                CREATE TABLE IF NOT EXISTS occurrences (
                    symbol TEXT NOT NULL,
                    kind TEXT NOT NULL,
                    path TEXT NOT NULL,
                    start_line INTEGER NOT NULL,
                    end_line INTEGER NOT NULL,
                    start_byte INTEGER NOT NULL,
                    end_byte INTEGER NOT NULL,
                    language TEXT,
                    container TEXT,
                    signature TEXT
                );
                CREATE TABLE IF NOT EXISTS files (
                    path TEXT PRIMARY KEY,
                    content_hash TEXT NOT NULL,
                    mtime_ns INTEGER NOT NULL,
                    language TEXT,
                    last_indexed TEXT NOT NULL
                );
                CREATE TABLE IF NOT EXISTS meta (
                    key TEXT PRIMARY KEY,
                    value TEXT NOT NULL
                );
                CREATE INDEX IF NOT EXISTS idx_occ_symbol_kind ON occurrences (symbol, kind);
                CREATE INDEX IF NOT EXISTS idx_occ_path ON occurrences (path);
                CREATE INDEX IF NOT EXISTS idx_occ_path_line ON occurrences (path, start_line);
                CREATE INDEX IF NOT EXISTS idx_occ_kind ON occurrences (kind);
                """)
            conn.execute(
                """
                INSERT INTO meta (key, value) VALUES (?, ?)
                ON CONFLICT(key) DO UPDATE SET value = excluded.value
                """,
                (SCHEMA_VERSION_KEY, SYMBOL_INDEX_SCHEMA_VERSION),
            )

    def _schema_problem(self) -> str | None:
        try:
            with self._connect() as conn:
                existing_tables = self._list_tables(conn)
                missing_tables = sorted(REQUIRED_TABLES - existing_tables)
                if missing_tables:
                    return f"required tables missing ({', '.join(missing_tables)})"
                for table, expected_columns in REQUIRED_COLUMNS.items():
                    existing_columns = self._list_columns(conn, table)
                    missing_columns = sorted(expected_columns - existing_columns)
                    if missing_columns:
                        return f"table '{table}' missing columns ({', '.join(missing_columns)})"
                version = self._read_meta_value(conn, SCHEMA_VERSION_KEY)
                if version != SYMBOL_INDEX_SCHEMA_VERSION:
                    found = version if version is not None else "none"
                    return (
                        "schema version mismatch "
                        f"(found {found}, expected {SYMBOL_INDEX_SCHEMA_VERSION})"
                    )
        except sqlite3.DatabaseError as exc:
            return f"sqlite open/integrity error: {exc}"
        return None

    def _reset_database(self, reason: str) -> None:
        self.last_reset_reason = reason
        self._delete_database_artifacts()

    def _delete_database_artifacts(self) -> None:
        db_path = self.config.db_path
        for path in (db_path, Path(f"{db_path}-wal"), Path(f"{db_path}-shm")):
            if path.exists():
                path.unlink()

    def get_file(self, path: str) -> IndexedFile | None:
        if not self.available:
            return None
        with self._connect() as conn:
            row = conn.execute(
                """
                SELECT path, content_hash, mtime_ns, language, last_indexed
                FROM files
                WHERE path = ?
                """,
                (path,),
            ).fetchone()
        if row is None:
            return None
        return IndexedFile(
            path=str(row["path"]),
            content_hash=str(row["content_hash"]),
            mtime_ns=int(row["mtime_ns"]),
            language=str(row["language"]) if row["language"] is not None else None,
            last_indexed=datetime.fromisoformat(str(row["last_indexed"])),
        )

    def list_indexed_paths(self) -> list[str]:
        if not self.available:
            return []
        with self._connect() as conn:
            rows = conn.execute("SELECT path FROM files ORDER BY path").fetchall()
        return [str(row["path"]) for row in rows]

    def upsert_file(self, indexed_file: IndexedFile) -> None:
        if not self.available:
            return
        with self._connect() as conn:
            conn.execute(
                """
                INSERT INTO files (path, content_hash, mtime_ns, language, last_indexed)
                VALUES (?, ?, ?, ?, ?)
                ON CONFLICT(path) DO UPDATE SET
                    content_hash = excluded.content_hash,
                    mtime_ns = excluded.mtime_ns,
                    language = excluded.language,
                    last_indexed = excluded.last_indexed
                """,
                (
                    indexed_file.path,
                    indexed_file.content_hash,
                    indexed_file.mtime_ns,
                    indexed_file.language,
                    indexed_file.last_indexed.isoformat(),
                ),
            )

    def replace_file_occurrences(
        self,
        *,
        indexed_file: IndexedFile,
        occurrences: list[SymbolOccurrence],
    ) -> None:
        if not self.available:
            return
        rows = [
            (
                item.symbol,
                item.kind,
                item.path,
                item.start_line,
                item.end_line,
                item.start_byte,
                item.end_byte,
                item.language,
                item.container,
                item.signature,
            )
            for item in occurrences
        ]
        with self._connect() as conn:
            conn.execute("DELETE FROM occurrences WHERE path = ?", (indexed_file.path,))
            if rows:
                conn.executemany(
                    """
                    INSERT INTO occurrences (
                        symbol, kind, path, start_line, end_line, start_byte, end_byte,
                        language, container, signature
                    )
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                    """,
                    rows,
                )
            conn.execute(
                """
                INSERT INTO files (path, content_hash, mtime_ns, language, last_indexed)
                VALUES (?, ?, ?, ?, ?)
                ON CONFLICT(path) DO UPDATE SET
                    content_hash = excluded.content_hash,
                    mtime_ns = excluded.mtime_ns,
                    language = excluded.language,
                    last_indexed = excluded.last_indexed
                """,
                (
                    indexed_file.path,
                    indexed_file.content_hash,
                    indexed_file.mtime_ns,
                    indexed_file.language,
                    indexed_file.last_indexed.isoformat(),
                ),
            )

    def delete_file(self, path: str) -> None:
        if not self.available:
            return
        with self._connect() as conn:
            conn.execute("DELETE FROM occurrences WHERE path = ?", (path,))
            conn.execute("DELETE FROM files WHERE path = ?", (path,))

    def prune_missing_files(self, *, seen_paths: set[str], scope_prefix: str | None = None) -> int:
        if not self.available:
            return 0
        normalized_prefix = os.path.normpath(scope_prefix) if scope_prefix else None
        removed = 0
        for indexed_path in self.list_indexed_paths():
            if normalized_prefix and not os.path.normpath(indexed_path).startswith(
                normalized_prefix.rstrip(os.sep) + os.sep
            ):
                if os.path.normpath(indexed_path) != normalized_prefix:
                    continue
            if indexed_path in seen_paths:
                continue
            self.delete_file(indexed_path)
            removed += 1
        return removed

    def list_occurrences(
        self,
        *,
        path_prefixes: tuple[str, ...] = (),
        language: str | None = None,
    ) -> list[SymbolOccurrence]:
        if not self.available:
            return []
        query = (
            "SELECT symbol, kind, path, start_line, end_line, start_byte, end_byte, "
            "language, container, signature "
            "FROM occurrences WHERE 1=1"
        )
        params: list[object] = []
        if language:
            query += " AND language = ?"
            params.append(language)
        if path_prefixes:
            query += " AND ("
            path_clauses: list[str] = []
            for prefix in path_prefixes:
                normalized = prefix.rstrip("/\\")
                if not normalized:
                    continue
                path_clauses.append("path = ?")
                params.append(normalized)
                path_clauses.append("path LIKE ? ESCAPE '\\'")
                escaped = self._escape_like(normalized)
                params.append(f"{escaped}/%")
                params.append(f"{escaped}\\\\%")
                path_clauses.append("path LIKE ? ESCAPE '\\'")
            if not path_clauses:
                query += "1=1"
            else:
                query += " OR ".join(path_clauses)
            query += ")"
        query += " ORDER BY path, start_line, end_line, symbol, kind"
        with self._connect() as conn:
            rows = conn.execute(query, params).fetchall()
        return [
            SymbolOccurrence(
                symbol=str(row["symbol"]),
                kind=str(row["kind"]),
                path=str(row["path"]),
                start_line=int(row["start_line"]),
                end_line=int(row["end_line"]),
                start_byte=int(row["start_byte"]) if row["start_byte"] is not None else None,
                end_byte=int(row["end_byte"]) if row["end_byte"] is not None else None,
                language=str(row["language"]) if row["language"] is not None else None,
                container=str(row["container"]) if row["container"] is not None else None,
                signature=str(row["signature"]) if row["signature"] is not None else None,
            )
            for row in rows
        ]

    @staticmethod
    def _escape_like(value: str) -> str:
        return value.replace("\\", "\\\\").replace("%", "\\%").replace("_", "\\_")

    @staticmethod
    def _list_tables(conn: sqlite3.Connection) -> set[str]:
        rows = conn.execute("""
            SELECT name
            FROM sqlite_master
            WHERE type = 'table' AND name NOT LIKE 'sqlite_%'
            """).fetchall()
        return {str(row[0]) for row in rows}

    @staticmethod
    def _list_columns(conn: sqlite3.Connection, table: str) -> set[str]:
        rows = conn.execute(f"PRAGMA table_info({table})").fetchall()
        return {str(row[1]) for row in rows}

    @staticmethod
    def _read_meta_value(conn: sqlite3.Connection, key: str) -> str | None:
        row = conn.execute("SELECT value FROM meta WHERE key = ?", (key,)).fetchone()
        if row is None:
            return None
        return str(row["value"])
